from __future__ import annotations
import numpy as np
from numpy.typing import NDArray

from acryo.backend import Backend, AnyArray

UPSAMPLE = 20


def upsample(
    res: AnyArray,
    res_ori: AnyArray,
    max_shifts: tuple[float, float, float],
    pad_width_eff: tuple[float, float, float],
    backend: Backend,
) -> tuple[NDArray[np.float32], float]:
    maxima = backend.asnumpy(backend.unravel_index(backend.argmax(res), res.shape))
    midpoints = np.asarray(res.shape, dtype=np.int32) // 2
    coords, local_offset = _create_mesh(
        maxima,
        max_shifts,
        midpoints.astype(np.float32),
        pad_width_eff,
        backend,
    )
    local_response = backend.map_coordinates(
        res_ori, coords, order=3, mode="constant", cval=-1.0, prefilter=True
    )
    local_maxima = backend.asnumpy(
        backend.unravel_index(backend.argmax(local_response), local_response.shape)
    )
    corr = backend.asnumpy(local_response[tuple(local_maxima)])
    loc_shift = local_maxima / UPSAMPLE + local_offset
    shifts = maxima - midpoints + loc_shift
    return shifts, corr  # type: ignore


def _create_mesh(
    maxima: NDArray[np.intp],
    max_shifts: tuple[float, float, float],
    midpoints: NDArray[np.float32],
    pad_width_eff: tuple[float, float, float],
    backend: Backend,
):
    """
    Create a 3 pixel x 3 pixel (if not upsampled) mesh for image upsampling.
    """
    shifts = np.asarray(maxima, dtype=np.float32) - midpoints
    _max_shifts = np.asarray(max_shifts, dtype=np.float32)
    left = -shifts - _max_shifts
    right = -shifts + _max_shifts
    local_shifts = [
        [
            int(round(max(float(shiftl), -1.0) * UPSAMPLE)),
            int(round(min(float(shiftr), 1.0) * UPSAMPLE)),
        ]
        for shiftl, shiftr in zip(left, right)
    ]
    mesh = backend.stack(
        backend.meshgrid(
            *[
                backend.arange(s0, s1 + 1) / UPSAMPLE + m + w
                for (s0, s1), m, w in zip(local_shifts, maxima, pad_width_eff)
            ],  # type: ignore
            indexing="ij",
        ),
        axis=0,
    )
    offset = backend.array([s0 for s0, s1 in local_shifts], dtype=np.float32) / UPSAMPLE
    return mesh, offset



# pyright: reportPrivateImportUsage=false
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import (
    Any,
    List,
    Callable,
    Generic,
    Iterable,
    TypeVar,
    TYPE_CHECKING,
    Iterator,
    Sequence,
    MutableSequence,
    overload,
)
from typing_extensions import ParamSpec, Self, Concatenate

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from dask import array as da
    from dask.delayed import Delayed

_P = ParamSpec("_P")
_R = TypeVar("_R")
_D = TypeVar("_D", bound=np.generic)


class DaskTaskIterator(Generic[_R]):
    def __init__(self, tasks: Iterable[da.Array | Delayed]) -> None:
        self._iter = tasks

    def __iter__(self) -> Iterator[da.Array | Delayed]:
        return iter(self._iter)

    def tolist(self) -> DaskTaskList[_R]:
        return DaskTaskList(self._iter)


class _DaskComputable(Generic[_R], ABC):
    @abstractmethod
    def _as_dask_list(self) -> list[Any]:
        """Convert to a list that is ready for dask computation"""

    def compute(self) -> list[_R]:
        from dask import compute

        return compute(self._as_dask_list())[0]


class DaskTaskList(_DaskComputable[_R]):
    def __init__(self, tasks: Iterable[da.Array | Delayed]) -> None:
        self._tasks = list(tasks)

    def count(self) -> int:
        return len(self._tasks)

    def _as_dask_list(self) -> list[Any]:
        return self._tasks

    def asarrays(self, shape: tuple[int, ...], dtype: type[_D]) -> DaskArrayList[_D]:
        from dask import array as da

        return DaskArrayList(
            da.from_delayed(task, shape=shape, dtype=dtype) for task in self._tasks
        )

    def tostack(self, shape: tuple[int, ...], dtype: Any, axis: int = 0) -> da.Array:
        from dask import array as da

        return da.stack(self.asarrays(shape, dtype), axis=axis)

    def __iter__(self) -> Iterator[da.Array | Delayed]:
        return iter(self._tasks)

    def __len__(self) -> int:
        return len(self._tasks)


class DaskTaskPool(DaskTaskList[_R], Generic[_P, _R]):
    def __init__(self, func: Delayed) -> None:
        self._func = func
        self._tasks: list[da.Array | Delayed] = []

    @classmethod
    def from_func(cls, func: Callable[_P, _R]) -> Self:
        from dask import delayed

        return cls(delayed(func))

    def add_task(self, *args: _P.args, **kwargs: _P.kwargs) -> Self:
        self._tasks.append(self._func(*args, **kwargs))
        return self

    def add_tasks(self, duplication: int, *args: _P.args, **kwargs: _P.kwargs) -> Self:
        task = self._func(*args, **kwargs)
        self._tasks.extend([task] * duplication)
        return self


class NestedDaskTaskList(
    MutableSequence[_DaskComputable[_R]], _DaskComputable[List[_R]]
):
    def __init__(self, tasks: Iterable[_DaskComputable[_R]]) -> None:
        self._tasks = list(tasks)

    def insert(self, index: int, task: _DaskComputable) -> None:
        return self._tasks.insert(index, task)

    def __getitem__(self, index: int) -> _DaskComputable:
        return self._tasks[index]

    def __setitem__(self, index: int, task: _DaskComputable) -> None:
        self._tasks[index] = task

    def __delitem__(self, index: int) -> None:
        del self._tasks[index]

    def __iter__(self) -> Iterator[_DaskComputable]:
        return iter(self._tasks)

    def __len__(self) -> int:
        return len(self._tasks)

    def _as_dask_list(self) -> list[Any]:
        return [t._as_dask_list() for t in self._tasks]


class DaskArrayList(Sequence["da.Array"], _DaskComputable[NDArray[_D]]):
    def __init__(self, arrays: Iterable[da.Array]):
        self._arrays = list(arrays)

    def __len__(self) -> int:
        return len(self._arrays)

    def __iter__(self) -> Iterator[da.Array]:
        return iter(self._arrays)

    @overload
    def __getitem__(self, index: int) -> da.Array:
        ...

    @overload
    def __getitem__(self, index: slice) -> list[da.Array]:
        ...

    def __getitem__(self, index):
        return self._arrays[index]

    @classmethod
    def concat(cls, obj: Iterable[Iterable[da.Array]]) -> Self:
        import itertools

        return cls(itertools.chain(*obj))

    def _as_dask_list(self) -> list[Any]:
        return self._arrays

    def as_stack(self, axis: int = 0) -> da.Array:
        from dask import array as da

        return da.stack(self, axis=axis)

    def map(
        self,
        func: Callable[Concatenate[da.Array, _P], _R],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> DaskTaskPool[[da.Array], _R]:
        pool = DaskTaskPool.from_func(func)
        [pool.add_task(s, *args, **kwargs) for s in self]
        return pool

    def map_blocks(self, func, *args, **kwargs):
        return DaskArrayList(
            [a.map_blocks(func, *args, **kwargs) for a in self._arrays]
        )

    def enumerate(self) -> Iterator[tuple[int, da.Array]]:
        return enumerate(self)


_R1 = TypeVar("_R1")
_R2 = TypeVar("_R2")


@overload
def compute(arg: _DaskComputable[_R]) -> list[_R]:
    ...


@overload
def compute(
    arg: tuple[_DaskComputable[_R], _DaskComputable[_R1]]
) -> tuple[list[_R], list[_R1]]:
    ...


@overload
def compute(
    arg: tuple[_DaskComputable[_R], _DaskComputable[_R1], _DaskComputable[_R2]]
) -> tuple[list[_R], list[_R1], list[_R2]]:
    ...


@overload
def compute(
    arg: Sequence[_DaskComputable[_R]],
) -> list[list[_R]]:
    ...


def compute(
    arg: _DaskComputable | tuple[_DaskComputable, ...] | Sequence[_DaskComputable]
):
    """Compute dask tasks"""
    from dask import compute

    if isinstance(arg, _DaskComputable):
        return arg.compute()

    elif isinstance(arg, tuple):
        return tuple(compute([a._as_dask_list() for a in arg])[0])

    elif isinstance(arg, list):
        return compute([a._as_dask_list() for a in arg])[0]

    else:
        raise TypeError(f"Invalid type: {type(arg)}")



from __future__ import annotations
from functools import lru_cache
import numpy as np
from numpy.typing import NDArray
from scipy.fft import next_fast_len

from acryo._types import pixel
from acryo.backend import Backend, AnyArray
from ._upsample import upsample

# Normalized cross correlation
UPSAMPLE = 20


def ncc_landscape(
    img0: AnyArray[np.float32],
    img1: AnyArray[np.float32],
    max_shifts: tuple[float, ...],
    backend: Backend,
) -> AnyArray[np.float32]:
    if max_shifts is not None:
        max_shifts = tuple(max_shifts)
    pad_width = _get_padding_width(max_shifts)
    padimg = backend.pad(img0, pad_width=pad_width, mode="constant", constant_values=0)
    return ncc_landscape_no_pad(padimg, img1, backend=backend)


def ncc_landscape_no_pad(
    img0: AnyArray[np.float32],
    img1: AnyArray[np.float32],
    backend: Backend,
) -> AnyArray[np.float32]:
    corr = fftconvolve(img0, img1[::-1, ::-1, ::-1], backend)[1:-1, 1:-1, 1:-1]

    win_sum1 = _window_sum_3d(img0, img1.shape, backend=backend)
    win_sum2 = _window_sum_3d(img0**2, img1.shape, backend=backend)

    template_mean = img1.mean()
    template_volume = np.prod(img1.shape, dtype=np.float32)  # type: ignore
    template_ssd = backend.sum((img1 - template_mean) ** 2)

    var = (win_sum2 - win_sum1**2 / template_volume) * template_ssd

    # zero division happens when perfectly matched
    response = backend.zeros(corr.shape, dtype=corr.dtype)
    mask = var > 0
    response[mask] = (corr - win_sum1 * template_mean)[mask] / _safe_sqrt(
        var, fill=np.inf, backend=backend
    )[mask]
    return response


def zncc_landscape_with_crop(
    img0: AnyArray[np.float32],
    img1: AnyArray[np.float32],
    max_shifts: tuple[float, ...],
    backend: Backend,
) -> AnyArray[np.float32]:
    response = ncc_landscape(
        img0 - img0.mean(), img1 - img1.mean(), max_shifts, backend=backend
    )
    pad_width_eff = tuple(
        (s - int(m) * 2 - 1) // 2 for m, s in zip(max_shifts, response.shape)
    )
    sl_res = tuple(slice(w, -w, None) for w in pad_width_eff)
    return response[sl_res]


def subpixel_zncc(
    img0: AnyArray[np.float32],
    img1: AnyArray[np.float32],
    max_shifts: pixel | tuple[pixel, ...],
    backend: Backend,
) -> tuple[NDArray[np.float32], float]:
    """Get the optimal subpixel shift and the score."""
    img0 = img0 - img0.mean()
    img1 = img1 - img1.mean()
    if isinstance(max_shifts, (int, float)):
        max_shifts = (max_shifts,) * img0.ndim
    response = ncc_landscape(img0, img1, max_shifts, backend=backend)
    pad_width_eff = tuple(
        (s - int(m) * 2 - 1) // 2 for m, s in zip(max_shifts, response.shape)
    )
    sl_res = tuple(slice(w, -w, None) for w in pad_width_eff)
    response_center = response[sl_res]
    return upsample(
        response_center, response, max_shifts, pad_width_eff, backend=backend
    )


def _window_sum_2d(
    image: AnyArray[np.float32],
    window_shape: tuple[int, ...],
    backend: Backend,
) -> AnyArray[np.float32]:
    window_sum = backend.cumsum(image, axis=0)
    window_sum = window_sum[window_shape[0] : -1] - window_sum[: -window_shape[0] - 1]
    window_sum = backend.cumsum(window_sum, axis=1)
    window_sum = (
        window_sum[:, window_shape[1] : -1] - window_sum[:, : -window_shape[1] - 1]
    )

    return window_sum


def _window_sum_3d(
    image: AnyArray[np.float32],
    window_shape: tuple[int, ...],
    backend: Backend,
) -> AnyArray[np.float32]:
    window_sum = _window_sum_2d(image, window_shape, backend=backend)
    window_sum = backend.cumsum(window_sum, axis=2)
    window_sum = (
        window_sum[:, :, window_shape[2] : -1]
        - window_sum[:, :, : -window_shape[2] - 1]
    )

    return window_sum


def _safe_sqrt(a: AnyArray[np.float32], fill: float, backend: Backend):
    out = backend.full(a.shape, fill, dtype=np.float32)
    out = backend.zeros(a.shape, dtype=a.dtype)
    mask = a > 0
    out[mask] = backend.sqrt(a[mask])
    return out


@lru_cache(maxsize=12)
def _get_padding_width(max_shifts: tuple[int, ...]) -> list[tuple[int, int]]:
    pad_width: list[tuple[int, int]] = []
    for w in max_shifts:
        w_int = int(np.ceil(w + 3))
        pad_width.append((w_int, w_int))

    return pad_width


def fftconvolve(
    in1: AnyArray[np.float32],
    in2: AnyArray[np.float32],
    backend: Backend,
):
    s1 = in1.shape
    s2 = in2.shape

    # shape = in1.shape
    shape = [s1[i] + s2[i] - 1 for i in range(in1.ndim)]

    # convolve in the frequency domain
    fshape: tuple[int, int, int] = tuple(
        next_fast_len(shape[a], False) for a in range(in1.ndim)  # type: ignore
    )

    sp1 = backend.rfftn(in1, fshape)
    sp2 = backend.rfftn(in2, fshape)
    ret = backend.irfftn(sp1 * sp2, fshape)

    fslice = tuple([slice(sz) for sz in shape])
    ret = ret[fslice]

    return _apply_conv_mode(ret, s1, s2)


def _apply_conv_mode(
    ret: AnyArray[np.float32], s1: tuple[int, ...], s2: tuple[int, ...]
):
    shape_valid = np.array([s1[a] - s2[a] + 1 for a in range(ret.ndim)], dtype=np.int32)
    currshape = np.array(ret.shape)
    startind = (currshape - shape_valid) // 2
    endind = startind + shape_valid
    sl = [slice(startind[k], endind[k]) for k in range(len(endind))]
    return ret[tuple(sl)]



# pyright: reportPrivateImportUsage=false
from __future__ import annotations

from types import MappingProxyType
from typing import (
    Callable,
    Iterable,
    NamedTuple,
    Sequence,
    TYPE_CHECKING,
    Union,
    TypeVar,
)
import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation

from acryo import _utils
from acryo.molecules import Molecules, axes_to_rotator
from acryo.pipe._classes import ImageProvider
from acryo._types import nm, pixel
from acryo._dask import DaskTaskPool
from acryo._typed_scipy import spline_filter, affine_transform

if TYPE_CHECKING:
    from typing_extensions import Self, Literal
    from scipy.spatial.transform import Rotation
    import polars as pl

ColorType = Union[tuple[float, float, float], tuple[float, float, float, float]]


class Component(NamedTuple):
    """A component of tomogram."""

    molecules: Molecules
    image: NDArray[np.float32] | ImageProvider


class TomogramSimulator:
    """
    An object for tomogram simulation

    A TomogramSimulator contains set(s) of molecules and their corresponding density
    and additional information for tomogram generation. For instance, to simulate a
    tomogram with two sets of molecules, you can do

    >>> sim = TomogramSimulator()
    >>> sim.add_molecules(molecules_a, image_a, name="Molecule-A")
    >>> sim.add_molecules(molecules_b, image_b, name="Molecule-B")
    >>> img = sim.simulate(shape=(128, 128, 128))

    Parameters
    ----------
    order : int, default is 3
        Interpolation order for density image.
        - 0 = Nearest neighbor
        - 1 = Linear interpolation
        - 3 = Cubic interpolation
    scale : float, default is 1.0
        Scale of the pixel. This value is used to determine the position of the
        molecules.
    corner_safe : bool, default is False
        Not implemented yet.
    """

    def __init__(
        self,
        order: Literal[0, 1, 3] = 3,
        scale: nm = 1.0,
        corner_safe: bool = False,
    ) -> None:
        # check interpolation order
        if order not in (0, 1, 3):
            raise ValueError(
                f"The third argument 'order' must be 0, 1 or 3, got {order!r}."
            )
        self._order: Literal[0, 1, 3] = order

        # check scale
        self._scale = float(scale)
        if self._scale <= 0:
            raise ValueError("Negative scale is not allowed.")
        self._corner_safe = corner_safe

        self._components: dict[str, Component] = {}

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return f"<{clsname} object with {len(self._components)} components>"

    @property
    def components(self) -> MappingProxyType[str, Component]:
        """Return all the components in the tomogram."""
        return MappingProxyType(self._components)  # type: ignore

    @property
    def order(self) -> Literal[0, 1, 3]:
        """Return the interpolation order."""
        return self._order

    @property
    def scale(self) -> float:
        """The physical scale of the tomogram."""
        return self._scale

    @property
    def corner_safe(self) -> bool:
        return self._corner_safe

    def replace(
        self,
        order: Literal[0, 1, 3] | None = None,
        scale: float | None = None,
        corner_safe: bool | None = None,
    ) -> Self:
        """Return a new instance with different parameter(s)."""
        if order is None:
            order = self.order
        if scale is None:
            scale = self.scale
        if corner_safe is None:
            corner_safe = self.corner_safe
        out = self.__class__(
            order=order,
            scale=scale,
            corner_safe=corner_safe,
        )
        out._components = self._components.copy()
        return out

    def copy(self) -> Self:
        """Create a shallow copy of the loader."""
        return self.replace()

    def add_molecules(
        self,
        molecules: Molecules,
        image: NDArray[np.float32] | ImageProvider,
        *,
        name: str | None = None,
        overwrite: bool = False,
    ) -> Self:
        """
        Add molecules to the tomogram.

        Parameters
        ----------
        molecules : Molecules
            Molecules object that defines the coordinates and orientations of the
            input molecules.
        image : np.ndarray or ImageProvider
            Density image of the input molecules, or an image provider.
        name : str, optional
            Name of the molecules., by default None
        overwrite : bool, default is False
            If true, allow overwriting existing component.

        Returns
        -------
        TomogramSimulator
            Same object.
        """
        if name is None:
            name = f"component<{hex(id(molecules))}>"
        if not overwrite and name in self._components:
            raise ValueError(f"Component of name {name!r} already exists.")
        if isinstance(image, np.ndarray):
            if image.dtype != np.float32:
                image = image.astype(np.float32)
        elif not isinstance(image, ImageProvider):
            raise TypeError(
                f"The second argument must be an array or an ImageProvider, got {type(image)}."
            )
        self._components[name] = Component(molecules, image)
        return self

    def collect_molecules(self) -> Molecules:
        return Molecules.concat(comp.molecules for comp in self._components.values())

    def subset(self, names: str | Sequence[str]) -> Self:
        """
        Construct a simulator composed of a subset of the components.

        Parameters
        ----------
        names : str or sequence of str
            Component names that will be included in the subset.

        Returns
        -------
        TomogramSimulator
            New simulator object.
        """
        if isinstance(names, str):
            _names = {names}
        else:
            _names = set(names)
        new = type(self)(
            order=self.order, scale=self._scale, corner_safe=self.corner_safe
        )
        for name, comp in self._components.items():
            if name in _names:
                new._components[name] = comp
        return new

    def simulate(
        self,
        shape: tuple[pixel, pixel, pixel],
        colormap: Callable[[pl.DataFrame], ColorType]
        | NDArray[np.float32]
        | None = None,
    ) -> NDArray[np.float32]:
        """
        Simulate a tomogram.

        Parameters
        ----------
        shape : tuple of int
            Shape of the tomogram.
        colormap: callable, optional
            Colormap used to generate the colored tomogram. The input is a polars
            row vector of features of each molecule.

        Returns
        -------
        np.ndarray
            Simulated tomogram image.
        """
        if len(self._components) == 0:
            import warnings

            warnings.warn("No molecules are added to the simulator.", UserWarning)

        if colormap is not None:
            return self._simulate_with_color(shape, colormap)
        if len(shape) != 3:
            raise ValueError("The shape must be a 3-tuple.")
        return self._simulate(shape)

    def _simulate(self, shape: tuple[pixel, pixel, pixel]):
        """Simulate a grayscale tomogram."""
        tomogram = np.zeros(shape, dtype=np.float32)
        pool = DaskTaskPool.from_func(_simulate_one)
        for mol, image in self._components.values():
            img = self._get_image(image)
            starts, stops, mtxs = _prep_iterators(mol, img.shape, self._scale)
            for start, stop, mtx in zip(starts, stops, mtxs):
                pool.add_task(img, start, stop, mtx, shape, self.order)

        results = pool.compute()
        for sl, img_fragment in results:
            if img_fragment is not None:
                tomogram[sl] += img_fragment

        return tomogram

    def _simulate_with_color(
        self,
        shape: tuple[pixel, pixel, pixel],
        colormap: Callable[[pl.DataFrame], ColorType] | NDArray[np.float32],
    ):
        """Simulate a colored tomogram."""
        if isinstance(colormap, np.ndarray):
            if len(self._components) != 1:
                raise ValueError(
                    "If an array is given as the colormap, cannot simulate colored tomogram"
                )
            if colormap.shape[1] not in (3, 4):
                raise ValueError(f"Wrong shape as a colormap {colormap.shape}.")
            cmap = lambda df, i: colormap[i]
        else:
            cmap = lambda df, i: colormap(df[i])

        tomogram = np.zeros((3,) + shape, dtype=np.float32)
        pool = DaskTaskPool.from_func(_simulate_color_one)
        for mol, image in self._components.values():
            img = self._get_image(image)
            starts, stops, mtxs = _prep_iterators(mol, img.shape, self._scale)
            feat = mol.features
            for i, (start, stop, mtx) in enumerate(zip(starts, stops, mtxs)):
                f0 = cmap(feat, i)
                # f0 = colormap(feat[i])
                pool.add_task(img, start, stop, mtx, shape, self.order, f0)

        results = pool.compute()
        for sl, img_fragment in results:
            if img_fragment is not None:
                tomogram[sl] += img_fragment

        return tomogram

    def simulate_2d(self, shape: tuple[pixel, pixel]) -> NDArray[np.float32]:
        """Simulate a grayscale tomogram."""
        projection = np.zeros(shape, dtype=np.float32)
        pool = DaskTaskPool.from_func(_simulate_2d_one)
        for mol, image in self._components.values():
            img = self._get_image(image)
            starts, stops, mtxs = _prep_iterators(mol, img.shape, self._scale)
            zsize = mol.pos[:, 0].max() / self.scale + np.sum(img.shape)
            shape3d = (int(np.ceil(zsize)),) + shape

            # reduce z axis
            starts, stops, mtxs = starts[1:], stops[1:], mtxs[1:]

            for start, stop, mtx in zip(starts, stops, mtxs):
                pool.add_task(img, start, stop, mtx, shape3d, self.order)

        results = pool.compute()
        for sl, img_fragment in results:
            if img_fragment is not None:
                projection[sl] += img_fragment

        return projection

    def simulate_projection(
        self,
        shape: tuple[int, int],
        center: tuple[float, float, float],
        xaxis: tuple[float, float, float],
        yaxis: tuple[float, float, float],
    ) -> NDArray[np.float32]:
        """
        Simulate a projection of the tomogram in any direction.

        Projection plane and the projection range are defined by the given parameters.
        ``shape`` and ``center`` define the range of the projection plane. Technically,
        ``center`` will be the scale-considered center of the assumed 3D tomogram image.
        ``xaxis`` and ``yaxis`` define the coordinate system of the projection plane.

        Parameters
        ----------
        shape : (int, int)
            Output shape of the projection.
        center : (float, float, float)
            Center of the projection plane in world coordinate system.
        xaxis : (float, float, float)
            X-axis of the projection plane in world coordinate system.
        yaxis : (float, float, float)
            Y-axis of the projection plane in world coordinate system.

        Returns
        -------
        2D array
            Simulated projection image.
        """
        projection = np.zeros(shape, dtype=np.float32)
        ex = np.asarray(xaxis, dtype=np.float32) / np.linalg.norm(xaxis)
        ey = np.asarray(yaxis, dtype=np.float32) / np.linalg.norm(yaxis)
        if np.abs(np.dot(ex, ey)) > 1e-6:
            raise ValueError(f"xaxis {xaxis!r} and yaxis {yaxis!r} are not orthogonal.")
        rc = np.asarray(center, dtype=np.float32)
        pool = DaskTaskPool.from_func(_simulate_projection_one)
        for mol, image in self._components.values():
            img = self._get_image(image)
            pos_scaled = (mol.pos - rc) / self.scale
            xcoords = pos_scaled.dot(ex) + (shape[1] - 1) / 2
            ycoords = pos_scaled.dot(ey) + (shape[0] - 1) / 2
            coords = np.stack((ycoords, xcoords), axis=1)
            glob_rotator = axes_to_rotator(cross(ex, ey), ey)
            for i, yx in enumerate(coords):
                pool.add_task(yx, shape, img, mol.rotator[i], glob_rotator)

        results = pool.compute()
        for sl, img_fragment in results:
            if img_fragment is not None:
                projection[sl] += img_fragment

        return projection

    def simulate_tilt_series(
        self,
        degrees: Iterable[float],
        shape: tuple[int, int, int],
    ) -> NDArray[np.float32]:
        """
        Simulate a tilt series of the tomogram.

        This method is a simplified version of :meth:`simulate_projection`. Projection
        angles are restricted to the rotation around the y-axis of the world coordinate
        and the tomogram will be projected parallel to the z-axis.

        Parameters
        ----------
        degrees : iterable of float
            Rotation angles.
        shape : (int, int, int)
            Tomogram shape.

        Returns
        -------
        (N, Y, X) array
            Tilting series image.
        """
        rads = np.deg2rad(list(degrees))
        ntilt = len(rads)
        ey = np.array([0, 1, 0], dtype=np.float32)
        rc = (np.array(shape, dtype=np.float32) / 2 - 0.5) * self.scale
        tilt_series = np.zeros((ntilt, shape[1], shape[2]), dtype=np.float32)
        pool = DaskTaskPool.from_func(_simulate_projection_one_labeled)
        for i, rad in enumerate(rads):
            ex = np.array([np.sin(rad), 0, np.cos(rad)], dtype=np.float32)
            for mol, image in self._components.values():
                img = self._get_image(image)
                pos_scaled = (mol.pos - rc) / self.scale
                xcoords = pos_scaled.dot(ex) + (shape[2] - 1) / 2
                ycoords = pos_scaled.dot(ey) + (shape[1] - 1) / 2
                coords = np.stack((ycoords, xcoords), axis=1)
                glob_rotator = axes_to_rotator(cross(ex, ey), ey)
                for ci, yx in enumerate(coords):
                    pool.add_task(yx, shape[1:], img, mol.rotator[ci], glob_rotator, i)

        results = pool.compute()
        for sl, img_fragment in results:
            if img_fragment is not None:
                tilt_series[sl] += img_fragment

        return tilt_series

    def _get_image(
        self, image: NDArray[np.float32] | ImageProvider
    ) -> NDArray[np.float32]:
        if isinstance(image, ImageProvider):
            img = image.provide(self.scale)
        else:
            img = image

        if self.order > 1:
            img = spline_filter(img, order=self.order, mode="constant")
        return img


def _compose_affine_matrices(
    center: np.ndarray,
    rotator: Rotation,
    output_center: np.ndarray | None = None,
) -> NDArray[np.float32]:
    """Compose Affine matrices from an array shape and a Rotation object."""
    if output_center is None:
        raise NotImplementedError

    dz, dy, dx = center

    # center to corner
    translation_0 = np.array(
        [
            [1.0, 0.0, 0.0, dz],
            [0.0, 1.0, 0.0, dy],
            [0.0, 0.0, 1.0, dx],
            [0.0, 0.0, 0.0, 1.0],
        ],
        dtype=np.float32,
    )
    # corner to center
    translation_1 = _eyes(len(rotator))
    translation_1[:, :3, 3] = -output_center

    rot_mtx = _eyes(len(rotator))
    rot_mtx[:, :3, :3] = rotator.as_matrix()

    return np.einsum("ij,njk,nkl->nil", translation_0, rot_mtx, translation_1)  # type: ignore


def _eyes(n: int) -> np.ndarray:
    return np.stack([np.eye(4, dtype=np.float32)] * n, axis=0)


def _prep_iterators(mol: Molecules, shape: tuple[int, int, int], scale: float):
    # image slice must be integer so split it into two parts
    pos = mol.pos / scale
    intpos = pos.astype(np.int32)
    residue = pos - intpos.astype(np.float32)

    # construct matrices
    center = (np.array(shape) - 1.0) / 2.0
    starts = intpos - center.astype(np.int32)
    stops = starts + shape
    mtxs = _compose_affine_matrices(
        center, mol.rotator.inv(), output_center=center + residue
    )

    return starts, stops, mtxs


_T = TypeVar("_T", bound=Iterable[int])


def _prep_slices(
    start: _T,
    stop: _T,
    tomogram_shape: _T,
    template_shape: _T,
) -> tuple[tuple[slice, ...], tuple[slice, ...] | None]:
    sl_src_list: list[slice] = []
    sl_dst_list: list[slice] = []
    for s, e, size, tsize in zip(start, stop, tomogram_shape, template_shape):
        try:
            _sl, _pads, _out_of_bound = _utils.make_slice_and_pad(s, e, size)
        except ValueError:
            # out-of-bound is not a problem while simulation
            return (slice(None),), None
        else:
            sl_dst_list.append(_sl)
            if _out_of_bound:
                s0, s1 = _pads
                sl_src_list.append(slice(s0, tsize - s1))
            else:
                sl_src_list.append(slice(None))
    sl_src = tuple(sl_src_list)
    sl_dst = tuple(sl_dst_list)
    return sl_src, sl_dst


def _simulate_one(
    img: NDArray[np.float32],
    start: tuple[int, int, int],
    stop: tuple[int, int, int],
    mtx: NDArray[np.float32],
    shape: tuple[int, int, int],
    order: int,
) -> tuple[tuple[slice, ...], NDArray[np.float32] | None]:
    sl_src, sl_dst = _prep_slices(start, stop, shape, img.shape)
    if sl_dst is None:
        return (slice(None),), None
    transformed = affine_transform(
        img,
        mtx,
        mode="constant",
        cval=0.0,
        order=order,
        prefilter=False,
    )
    return sl_dst, transformed[sl_src]


def _simulate_color_one(
    img: NDArray[np.float32],
    start: tuple[int, int, int],
    stop: tuple[int, int, int],
    mtx: NDArray[np.float32],
    shape: tuple[int, int, int],
    order: int,
    color: tuple[float, float, float] | tuple[float, float, float, float],
) -> tuple[tuple[slice, ...], NDArray[np.float32] | None]:
    if len(color) == 4:
        _cr, _cg, _cb, _alpha = color
    else:
        _cr, _cg, _cb = color
        _alpha = 1.0
    sl_src, sl_dst = _prep_slices(start, stop, shape, img.shape)
    if sl_dst is None:
        return (slice(None),), None

    sl_dst = (slice(None),) + sl_dst

    transformed = affine_transform(
        img,
        mtx,
        mode="constant",
        cval=0.0,
        order=order,
        prefilter=False,
    )
    img_min = img.min()
    img_max = img.max()
    _a = (transformed[sl_src] - img_min) / (img_max - img_min) * _alpha
    color_array = np.stack([_cr * _a, _cg * _a, _cb * _a], axis=0)

    return sl_dst, color_array


def _simulate_2d_one(
    img: NDArray[np.float32],
    start: tuple[int, int, int],
    stop: tuple[int, int, int],
    mtx: NDArray[np.float32],
    shape: tuple[int, int, int],
    order: int,
):
    sl_src, sl_dst = _prep_slices(start, stop, shape, img.shape)
    if sl_dst is None:
        return (slice(None),), None
    transformed = affine_transform(
        img,
        mtx,
        mode="constant",
        cval=0.0,
        order=order,
        prefilter=False,
    )
    projected = np.sum(transformed[sl_src], axis=0)
    return sl_dst[1:], projected


def _simulate_projection_one(
    yx: NDArray[np.float32],
    shape: tuple[int, int],
    image: NDArray[np.float32],
    rotator: Rotation,
    glob_rotator: Rotation,
) -> tuple[tuple[slice, slice], NDArray[np.float32] | None]:
    proj_shape = np.array(image.shape[1:], dtype=np.int32)
    min_ = yx - proj_shape.astype(np.float32) / 2 + 0.5
    int_min_ = min_.astype(np.int32)
    int_max_ = int_min_ + proj_shape

    sl_src, sl_dst = _prep_slices(int_min_, int_max_, shape, proj_shape)

    if sl_dst is None:
        return (slice(None), slice(None)), None

    center = np.array(image.shape) / 2 - 0.5
    residue = np.array([0, *(min_ - int_min_)])
    mtx = _compose_affine_matrices(
        center, rotator.inv() * glob_rotator, center + residue
    )[0]

    transformed = affine_transform(
        image, mtx, mode="constant", cval=0.0, order=3, prefilter=False
    )
    projection: NDArray[np.float32] = np.sum(transformed, axis=0)
    return sl_dst, projection[sl_src]


def _simulate_projection_one_labeled(
    yx: NDArray[np.float32],
    shape: tuple[int, int],
    image: NDArray[np.float32],
    rotator: Rotation,
    glob_rotator: Rotation,
    idx: int,
) -> tuple[tuple[int, slice, slice], NDArray[np.float32] | None]:
    sl, img = _simulate_projection_one(yx, shape, image, rotator, glob_rotator)
    return (idx,) + sl, img


def cross(x: np.ndarray, y: np.ndarray, axis=None) -> np.ndarray:
    """Vector outer product in zyx coordinate."""
    return -np.cross(x, y, axis=axis)  # type: ignore



from __future__ import annotations
from functools import lru_cache
import numpy as np
from numpy.typing import NDArray

from acryo.backend import Backend, AnyArray
from ._upsample import upsample


def fsc_landscape(
    ft0: AnyArray[np.complex64],
    ft1: AnyArray[np.complex64],
    max_shifts: tuple[float, float, float],
    backend: Backend,
) -> AnyArray[np.float32]:
    shape = ft0.shape
    labels = _get_radial_label(shape, backend)
    nlabels = int(labels.max())
    index = backend.arange(0, nlabels + 1)

    pw1 = ft1.real**2 + ft1.imag**2
    sigma1 = backend.sqrt(backend.sum_labels(pw1, labels=labels, index=index))
    out_shape = tuple(int(np.ceil(m)) * 2 + 1 for m in max_shifts)
    out = backend.zeros(out_shape)
    phase_z, phase_y, phase_x = _get_phases(shape, out_shape, backend)
    for iz, phiz in enumerate(phase_z):
        ft0_shifted_z = ft0 * phiz
        for iy, phiy in enumerate(phase_y):
            ft0_shifted_yz = ft0_shifted_z * phiy
            for ix, phix in enumerate(phase_x):
                ft0_shifted = ft0_shifted_yz * phix
                cov = ft0_shifted.real * ft1.real + ft0_shifted.imag * ft1.imag
                pw0 = ft0_shifted.real**2 + ft0_shifted.imag**2
                sigma0 = backend.sqrt(
                    backend.sum_labels(pw0, labels=labels, index=index)
                )
                fsc = backend.sum_labels(cov, labels=labels, index=index) / (
                    sigma0 * sigma1
                )
                out[iz, iy, ix] = float(fsc.mean())
    return out


def subpixel_fsc(
    ft0: AnyArray[np.complex64],
    ft1: AnyArray[np.complex64],
    max_shifts: tuple[float, float, float],
    backend: Backend,
) -> tuple[NDArray[np.float32], float]:
    out = fsc_landscape(ft0, ft1, max_shifts, backend=backend)
    return upsample(out, out, max_shifts, (0, 0, 0), backend=backend)


@lru_cache(maxsize=12)
def _get_radial_label(
    shape: tuple[int, int, int], backend: Backend
) -> AnyArray[np.uint16]:
    freqs = backend.meshgrid(*[backend.fftfreq(s) for s in shape], indexing="ij")
    r = backend.sqrt(sum(f**2 for f in freqs))  # type: ignore
    dfreq = 1.0 / min(shape)
    return (r / dfreq).astype(np.uint16)


def _get_phase_1d(
    mesh: AnyArray[np.float32], size: int, backend: Backend
) -> list[AnyArray[np.complex64]]:
    s = size // 2
    rng = range(-s, s + 1)
    return [backend.exp(2j * np.pi * x0 * mesh) for x0 in rng]  # type: ignore


@lru_cache(maxsize=12)
def _get_phases(shape, out_shape, backend: Backend):
    mesh = backend.meshgrid(*[backend.fftfreq(s) for s in shape], indexing="ij")
    phase_z = _get_phase_1d(mesh[0], out_shape[0], backend)
    phase_y = _get_phase_1d(mesh[1], out_shape[1], backend)
    phase_x = _get_phase_1d(mesh[2], out_shape[2], backend)
    return phase_z, phase_y, phase_x



from ._api import using_backend, set_backend, Backend, NUMPY_BACKEND, AnyArray
from ._mesh import build_mesh

__all__ = [
    "using_backend",
    "set_backend",
    "Backend",
    "NUMPY_BACKEND",
    "AnyArray",
    "build_mesh",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal
import numpy as np
from numpy.typing import NDArray

# scipy is not well typed. Make patches here.
if TYPE_CHECKING:
    Mode = Literal["constant", "nearest", "mirror", "wrap", "reflect"]

    def fftn(
        img: NDArray[np.float32] | NDArray[np.complex64], s=None, axes=None
    ) -> NDArray[np.complex64]:
        ...

    def ifftn(img: NDArray[np.complex64], s=None, axes=None) -> NDArray[np.complex64]:
        ...

    def rfftn(img: NDArray[np.float32], s=None, axes=None) -> NDArray[np.complex64]:
        ...

    def irfftn(img: NDArray[np.complex64], s=None, axes=None) -> NDArray[np.float32]:
        ...

    def convolve(
        in1: NDArray[np.float32],
        in2: NDArray[np.float32],
        mode: Literal["full", "valid", "same"] = "full",
        method: Literal["auto", "fft", "direct"] = "auto",
    ) -> NDArray[np.float32]:
        ...

    def spline_filter(
        input: NDArray[np.float32],
        order: int = 3,
        output: Any = np.float64,
        mode: Mode = "mirror",
    ) -> NDArray[np.float32]:
        ...

    def affine_transform(
        img: NDArray[np.float32],
        matrix: NDArray[np.float32],
        offset: float = 0.0,
        output_shape: tuple[int, ...] | None = None,
        output: NDArray[np.float32] | None = None,
        order: int = 3,
        mode: Mode = "constant",
        cval: float = 0.0,
        prefilter: bool = True,
    ) -> NDArray[np.float32]:
        ...

    def shift(
        input: NDArray[np.float32],
        shift: tuple[float, ...] | NDArray[np.float32],
        output: NDArray[np.float32] | None = None,
        order: int = 3,
        mode: Mode = "constant",
        cval: float = 0.0,
        prefilter: bool = True,
    ) -> NDArray[np.float32]:
        ...

    def zoom(
        input: NDArray[np.float32],
        zoom: float | tuple[float, ...] | NDArray[np.float32],
        output: NDArray[np.float32] | None = None,
        order: int = 3,
        mode: Mode = "constant",
        cval: float = 0.0,
        prefilter: bool = True,
    ) -> NDArray[np.float32]:
        ...

    def map_coordinates(
        input: NDArray[np.float32],
        coordinates: NDArray[np.float32],
        output: NDArray[np.float32] | None = None,
        order: int = 3,
        mode: Mode = "constant",
        cval: float = 0.0,
        prefilter: bool = True,
    ) -> NDArray[np.float32]:
        ...

    def sum_labels(
        arr: NDArray[np.float32],
        labels: NDArray[np.integer],
        index: NDArray[np.integer],
    ) -> NDArray[np.float32]:
        ...

else:
    from scipy.fft import rfftn, irfftn, fftn, ifftn
    from scipy.ndimage import (
        spline_filter,
        affine_transform,
        map_coordinates,
        shift,
        zoom,
        sum_labels,
    )
    from scipy.signal import convolve

__all__ = [
    "fftn",
    "ifftn",
    "rfftn",
    "irfftn",
    "spline_filter",
    "affine_transform",
    "shift",
    "zoom",
    "sum_labels",
    "map_coordinates",
    "convolve",
]



# Mostly copied from dask-ml
# See https://github.com/dask/dask-ml/blob/main/dask_ml/decomposition/pca.py

# pyright: reportPrivateImportUsage=false
from __future__ import annotations
import numbers

import dask
from dask import array as da, delayed
import numpy as np
from sklearn.decomposition import PCA
from sklearn.utils.extmath import fast_logdet
from sklearn.utils.validation import check_is_fitted, check_random_state


_TYPE_MSG = (
    "Got an unsupported type ({}). Dask-ML's PCA only support Dask Arrays or "
    "DataFrames.\n\nTo resolve this issue,\n\n"
    "  * Use Scikit-learn's PCA through `sklearn.decomposition.PCA`  # recommended\n\n"
    "Wrapping the input with a Dask Array/DataFrame will resolve "
    "this issue but is *not recommended* because Dask-ML's PCA "
    "implementation will likely be slower because the data fits in memory"
)


class DaskPCA(PCA):
    """Principal component analysis (PCA)
    Linear dimensionality reduction using Singular Value Decomposition of the
    data to project it to a lower dimensional space.
    It uses the "tsqr" algorithm from Benson et. al. (2013). See the References
    for more.
    Read more in the :ref:`User Guide <PCA>`.
    Parameters
    ----------
    n_components : int, or None
        Number of components to keep.
        if n_components is not set all components are kept::
            n_components == min(n_samples, n_features)
        .. note::
           Unlike scikit-learn, ``n_components='mle'`` and ``n_components``
           between ``(0, 1)`` are not currently supported.
    copy : bool (default True)
        ignored
    whiten : bool, optional (default False)
        When True (False by default) the `components_` vectors are multiplied
        by the square root of n_samples and then divided by the singular values
        to ensure uncorrelated outputs with unit component-wise variances.
        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometime
        improve the predictive accuracy of the downstream estimators by
        making their data respect some hard-wired assumptions.
    svd_solver : string {'auto', 'full', 'tsqr', 'randomized'}
        auto :
            the solver is selected by a default policy based on `X.shape` and
            `n_components`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient 'randomized'
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards.
        full :
            run exact full SVD and select the components by postprocessing
        randomized :
            run randomized SVD by using ``da.linalg.svd_compressed``.
    tol : float >= 0, optional (default .0)
        ignored
    iterated_power : int >= 0, default 0
        Number of iterations for the power method computed by
        svd_solver == 'randomized'.
    random_state : int, RandomState instance or None, optional (default None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `da.random`. Used when ``svd_solver`` == 'randomized'.
    Attributes
    ----------
    components_ : array, shape (n_components, n_features)
        Principal axes in feature space, representing the directions of
        maximum variance in the data. The components are sorted by
        ``explained_variance_``.
    explained_variance_ : array, shape (n_components,)
        The amount of variance explained by each of the selected components.
        Equal to n_components largest eigenvalues
        of the covariance matrix of X.
    explained_variance_ratio_ : array, shape (n_components,)
        Percentage of variance explained by each of the selected components.
        If ``n_components`` is not set then all components are stored and the
        sum of the ratios is equal to 1.0.
    singular_values_ : array, shape (n_components,)
        The singular values corresponding to each of the selected components.
        The singular values are equal to the 2-norms of the ``n_components``
        variables in the lower-dimensional space.
    mean_ : array, shape (n_features,)
        Per-feature empirical mean, estimated from the training set.
        Equal to `X.mean(axis=0)`.
    n_components_ : int
        The estimated number of components. When n_components is set
        to 'mle' or a number between 0 and 1 (with svd_solver == 'full') this
        number is estimated from input data. Otherwise it equals the parameter
        n_components, or the lesser value of n_features and n_samples
        if n_components is None.
    noise_variance_ : float
        The estimated noise covariance following the Probabilistic PCA model
        from Tipping and Bishop 1999. See "Pattern Recognition and
        Machine Learning" by C. Bishop, 12.2.1 p. 574 or
        http://www.miketipping.com/papers/met-mppca.pdf. It is required to
        computed the estimated data covariance and score samples.
        Equal to the average of (min(n_features, n_samples) - n_components)
        smallest eigenvalues of the covariance matrix of X.
    References
    ----------
    Direct QR factorizations for tall-and-skinny matrices in
    MapReduce architectures.
    A. Benson, D. Gleich, and J. Demmel.
    IEEE International Conference on Big Data, 2013.
    http://arxiv.org/abs/1301.1071
    Examples
    --------
    >>> import numpy as np
    >>> import dask.array as da
    >>> from dask_ml.decomposition import PCA
    >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    >>> dX = da.from_array(X, chunks=X.shape)
    >>> pca = PCA(n_components=2)
    >>> pca.fit(dX)
    PCA(copy=True, iterated_power='auto', n_components=2, random_state=None,
      svd_solver='auto', tol=0.0, whiten=False)
    >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [ 0.99244...  0.00755...]
    >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
    [ 6.30061...  0.54980...]
    >>> pca = PCA(n_components=2, svd_solver='full')
    >>> pca.fit(dX)                 # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    PCA(copy=True, iterated_power='auto', n_components=2, random_state=None,
      svd_solver='full', tol=0.0, whiten=False)
    >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [ 0.99244...  0.00755...]
    >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
    [ 6.30061...  0.54980...]
    Notes
    -----
    Differences from scikit-learn:
    * svd_solver : 'randomized' uses ``dask.linalg.svd_compressed``
      'full' uses ``dask.linalg.svd``, 'arpack' is not valid.
    * iterated_power : defaults to ``0``, the default for
      ``dask.linalg.svd_compressed``.
    * n_components : ``n_components='mle'`` is not allowed.
      Fractional ``n_components`` between 0 and 1 is not allowed.
    """

    def __init__(
        self,
        n_components=None,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power=0,
        random_state=None,
    ):
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state

    def fit(self, X: da.Array, y=None):
        if not dask.is_dask_collection(X):
            raise TypeError(_TYPE_MSG.format(type(X)))
        self._fit(X)
        self.n_features_in_ = X.shape[1]
        return self

    def _get_solver(self, X: da.Array, n_components: int):
        n_samples, n_features = X.shape
        solvers = {"full", "auto", "tsqr", "randomized"}
        solver = self.svd_solver

        if solver not in solvers:
            raise ValueError(
                "Invalid solver '{}'. Must be one of {}".format(solver, solvers)
            )

        if solver == "auto":
            # Small problem, just call full PCA
            if not _known_shape(X.shape):
                raise ValueError(
                    "Cannot automatically choose PCA solver with unknown "
                    "shapes. To clear this error,\n\n"
                    "    * pass X.to_dask_array(lengths=True)  "
                    "# for Dask DataFrame (dask >= 0.19)\n"
                    "    * pass X.compute_chunk_sizes()  "
                    "# for Dask Array X (dask >= 2.4)\n"
                    "    * Use a specific SVD solver "
                    "(e.g., ensure `svd_solver in ['randomized', 'tsqr', 'full']`)"
                )
            if max(n_samples, n_features) <= 500:
                solver = "full"
            elif n_components >= 1 and n_components < 0.8 * min(n_samples, n_features):
                solver = "randomized"
            # This is also the case of n_components in (0,1)
            else:
                solver = "full"

        if solver == "randomized":
            lower_limit = 1
        else:
            lower_limit = 0

        if not (np.nanmin([n_samples, n_features]) >= n_components >= lower_limit):
            msg = (
                "n_components={} must be between {} and "
                "min(n_samples, n_features)={} with "
                "svd_solver='{}'.".format(
                    n_components, lower_limit, min(n_samples, n_features), solver
                )
            )
            raise ValueError(msg)
        return solver

    def _fit(self, X: da.Array):
        # Handle n_components==None
        if self.n_components is None:
            n_components = min(X.shape)
        elif 0 < self.n_components < 1:
            raise NotImplementedError(
                "Fractional 'n_components' is not " "currently supported"
            )
        else:
            n_components = self.n_components

        solver = self._get_solver(X, n_components)

        self.mean_ = X.mean(0)
        X -= self.mean_

        if solver in {"full", "tsqr"}:
            U, S, V = da.linalg.svd(X)
        else:
            # randomized
            random_state = check_random_state(self.random_state)
            seed = draw_seed(random_state, np.iinfo("int32").max)
            n_power_iter = self.iterated_power
            U, S, V = da.linalg.svd_compressed(
                X, n_components, n_power_iter=n_power_iter, seed=seed
            )
        components, singular_values = V, S

        if solver == "randomized":
            total_variance = X.var(ddof=1, axis=0).sum()
        else:
            total_variance = np.nan

        try:
            (
                (self.n_samples_, self.n_features_in_),
                self.n_components_,
                self.components_,
                self.singular_values_,
                total_variance,
            ) = da.compute(
                delayed(lambda x: x.shape)(X),
                n_components,
                components,
                singular_values,
                total_variance,
            )
        except ValueError as e:
            if np.isnan(X.shape).any():
                msg = (
                    "Computation of the SVD raised an error. It is possible "
                    "n_components is too large. i.e., "
                    "`n_components > np.nanmin(X.shape) = "
                    "np.nanmin({})`\n\n"
                    "A possible resolution to this error is to ensure that "
                    "n_components <= min(n_samples, n_features)"
                )
                raise ValueError(msg.format(X.shape)) from e
            raise e

        self.explained_variance_ = self.singular_values_**2 / (self.n_samples_ - 1)

        if solver == "randomized":
            self.explained_variance_ratio_ = self.explained_variance_ / total_variance
        else:
            self.explained_variance_ratio_ = (
                self.explained_variance_ / self.explained_variance_.sum()
            )

        # Compute noise covariance using Probabilistic PCA model
        # The sigma2 maximum likelihood (cf. eq. 12.46)
        self.noise_variance_ = 0.0
        if n_components < min(self.n_features_in_, self.n_samples_):
            if solver == "randomized":
                self.noise_variance_ = (
                    total_variance - self.explained_variance_.sum()
                ) / (min(self.n_features_in_, self.n_samples_) - n_components)
            else:
                if n_components < len(self.explained_variance_):
                    self.noise_variance_ = self.explained_variance_[
                        n_components:
                    ].mean()

        self.components_ = self.components_[:n_components]
        self.explained_variance_ = self.explained_variance_[:n_components]
        self.explained_variance_ratio_ = self.explained_variance_ratio_[:n_components]
        self.singular_values_ = self.singular_values_[:n_components]

        if len(self.singular_values_) < n_components:
            self.n_components_ = len(self.singular_values_)
            msg = (
                "n_components={n} is larger than the number of singular values"
                " ({s}) (note: PCA has attributes as if n_components == {s})"
            )
            raise ValueError(msg.format(n=n_components, s=len(self.singular_values_)))

        return U, S, V

    def transform(self, X: da.Array) -> da.Array:
        """Apply dimensionality reduction on X.
        X is projected on the first principal components previous extracted
        from a training set.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples in the number of samples
            and n_features is the number of features.
        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
        """
        check_is_fitted(self, ["mean_", "components_"])

        # X = check_array(X)
        if self.mean_ is not None:
            X = X - self.mean_
        X_transformed = da.dot(X, self.components_.T)
        if self.whiten:
            X_transformed /= np.sqrt(self.explained_variance_)
        return X_transformed

    def fit_transform(self, X, y=None):
        """Fit the model with X and apply the dimensionality reduction on X.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples in the number of samples
            and n_features is the number of features.
        y : Ignored
        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
        """
        if not dask.is_dask_collection(X):
            raise TypeError(_TYPE_MSG.format(type(X)))
        U, S, V = self._fit(X)
        U = U[:, : self.n_components_]

        if self.whiten:
            # X_new = X * V / S * sqrt(n_samples) = U * sqrt(n_samples)
            U *= np.sqrt(X.shape[0] - 1)
        else:
            # X_new = X * V = U * S * V^T * V = U * S
            U *= S[: self.n_components_]

        return U

    def inverse_transform(self, X):
        """Transform data back to its original space.
        Returns an array X_original whose transform would be X.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_components)
            New data, where n_samples in the number of samples
            and n_components is the number of components.
        Returns
        -------
        X_original array-like, shape (n_samples, n_features)
        Notes
        -----
        If whitening is enabled, inverse_transform does not compute the
        exact inverse operation of transform.
        """
        check_is_fitted(self, "mean_")

        if self.whiten:
            return (
                da.dot(
                    X,
                    np.sqrt(self.explained_variance_[:, np.newaxis]) * self.components_,
                )
                + self.mean_
            )
        else:
            return da.dot(X, self.components_) + self.mean_

    def score_samples(self, X):
        """Return the log-likelihood of each sample.
        See. "Pattern Recognition and Machine Learning"
        by C. Bishop, 12.2.1 p. 574
        or http://www.miketipping.com/papers/met-mppca.pdf
        Parameters
        ----------
        X : array, shape(n_samples, n_features)
            The data.
        Returns
        -------
        ll : array, shape (n_samples,)
            Log-likelihood of each sample under the current model
        """
        check_is_fitted(self, "mean_")

        # X = check_array(X)
        Xr = X - self.mean_
        n_features = X.shape[1]
        precision = self.get_precision()  # [n_features, n_features]
        log_like = -0.5 * (Xr * (da.dot(Xr, precision))).sum(axis=1)
        log_like -= 0.5 * (n_features * da.log(2.0 * np.pi) - fast_logdet(precision))
        return log_like

    def score(self, X, y=None):
        """Return the average log-likelihood of all samples.
        See. "Pattern Recognition and Machine Learning"
        by C. Bishop, 12.2.1 p. 574
        or http://www.miketipping.com/papers/met-mppca.pdf
        Parameters
        ----------
        X : array, shape(n_samples, n_features)
            The data.
        y : Ignored
        Returns
        -------
        ll : float
            Average log-likelihood of the samples under the current model
        """
        return da.mean(self.score_samples(X))


def _known_shape(shape):
    return all(isinstance(x, numbers.Integral) for x in shape)


def draw_seed(random_state, low, high=None, size=None, dtype=None, chunks=None):
    kwargs = {"size": size}
    if chunks is not None:
        kwargs["chunks"] = chunks

    seed = random_state.randint(low, high, **kwargs)
    if dtype is not None and isinstance(seed, (da.Array, np.ndarray)):
        seed = seed.astype(dtype)

    return seed



# pyright: reportPrivateImportUsage=false
from __future__ import annotations

from ._base import LoaderBase
from typing import TYPE_CHECKING, Iterable, Sequence
import numpy as np
from numpy.typing import NDArray
from dask import array as da, delayed
from scipy import ndimage as ndi
from scipy.spatial.transform import Rotation

from acryo import _utils
from acryo._types import nm
from acryo._typed_scipy import fftn, ifftn, spline_filter
from acryo._dask import DaskArrayList, DaskTaskPool
from acryo.molecules import Molecules
from acryo.backend import Backend
from acryo.pipe._classes import ImageProvider

if TYPE_CHECKING:
    from typing_extensions import Self
    from ._base import _ShapeType


class MockLoader(LoaderBase):
    """
    A subtomogram loader from a virtual tomogram.

    This loader is for testing purpose only. The tomogram does not actually exist
    but subtomograms are generated on the fly based on the template image. The
    true center of the molecules is always at (0, 0, 0) and the true rotation is
    always the identity rotation. Thus, given molecule positions/rotations will
    directly be the values to be tested.

    Parameters
    ----------
    template : 3D array, ImageProvider
        The template image.
    molecules : Molecules
        The molecules to be loaded. Note that the true center/rotation is always
        (0, 0, 0) and the identity rotation, respectively.
    noise : float, optional
        Standard deviation of the Gaussian noise to be added to the subtomograms
        projection slices.
    degrees : array-like, optional
        The rotation angles in degrees. If not provided, projection/back-projection
        will not be simulated.
    central_axis : tuple, default is (0.0, 1.0, 0.0)
        The central axis of the rotation during tomogram acquisition.
    """

    def __init__(
        self,
        template: NDArray[np.float32] | ImageProvider,
        molecules: Molecules,
        noise: float = 0.0,
        degrees: Sequence[float] | NDArray[np.float32] | None = None,
        central_axis: tuple[float, float, float] = (0.0, 1.0, 0.0),
        order: int = 3,
        scale: nm = 1.0,
        corner_safe: bool = False,
    ) -> None:
        if noise < 0:
            raise ValueError("Noise must be non-negative.")
        super().__init__(order, scale, corner_safe)
        self._template = template
        self._noise = noise
        if degrees is None:
            if noise > 0:
                # TODO: implement this
                raise NotImplementedError(
                    "Noise simulation without degrees is not implemented yet."
                )
            self._degrees = None
        else:
            self._degrees = np.asarray(degrees, dtype=np.float32)
        self._molecules = molecules
        self._central_axis = np.array(central_axis, dtype=np.float32)
        if isinstance(template, ImageProvider):
            self._output_shape = template(self.scale).shape
        else:
            self._output_shape = template.shape
        if len(self._output_shape) != 3:
            raise ValueError("Template must be 3D.")

    @property
    def molecules(self) -> Molecules:
        """All the molecules"""
        return self._molecules

    def construct_loading_tasks(
        self,
        output_shape: _ShapeType = None,
        backend: Backend | None = None,
    ) -> DaskArrayList:
        # TODO: this implementation is not efficent. Radon transformation is not
        # actually needed. Apply missing wedge mask directly to the template, and
        # apply inverse-Radon only to the noise. The linearity of Radon
        # transformation guarantees that the result is correct.
        _backend = backend or Backend()
        if _backend.name == "cupy":
            raise NotImplementedError("Cupy backend is not supported yet.")
        if isinstance(self._template, ImageProvider):
            template = self._template(self.scale)
        else:
            template = self._template
        if output_shape is None:
            output_shape = template.shape
        elif output_shape != template.shape:
            raise ValueError(
                f"Mismatched output shape {output_shape!r} and template shape "
                f"{template.shape}."
            )

        # spline prefilter in advance
        if self.order > 1:
            template = spline_filter(
                template, order=self.order, mode="constant", output=np.float32
            )
        center = np.array(template.shape) / 2 - 0.5
        matrices = self.molecules.affine_matrix(
            center, center + self.molecules.pos / self.scale
        )
        pool = DaskTaskPool.from_func(_backend.affine_transform)
        for mtx in matrices:
            pool.add_task(template, mtx, order=self.order, prefilter=False)
        task_list = pool.asarrays(shape=template.shape, dtype=np.float32)
        if self._degrees is not None:
            task_list = DaskArrayList(
                simulate_noise(
                    task,
                    self._central_axis,
                    self._degrees,
                    self._noise,
                    seed=i,
                )
                for i, task in task_list.enumerate()
            )

        return task_list

    def replace(
        self,
        molecules: Molecules | None = None,
        output_shape: None = None,  # just for compatibility
        order: int | None = None,
        scale: float | None = None,
        corner_safe: bool | None = None,
    ) -> Self:
        if molecules is None:
            molecules = self.molecules
        if order is None:
            order = self.order
        if scale is None:
            scale = self.scale
        if corner_safe is None:
            corner_safe = self.corner_safe
        return self.__class__(
            self._template,
            molecules=molecules,
            noise=self._noise,
            degrees=self._degrees,
            order=order,
            scale=scale,
            corner_safe=corner_safe,
        )


def simulate_noise(
    img: NDArray[np.float32] | da.Array,
    central_axis,
    degrees: NDArray[np.float32],
    noise: float,
    seed: int,
) -> da.Array:
    # img: spline filtered
    matrices, output_shape = normalize_radon_input(img.shape, central_axis, degrees)
    sino: da.Array = da.stack(
        [
            da.from_delayed(
                radon_single(img, mtx, order=3, output_shape=output_shape),
                shape=output_shape[1:],
                dtype=np.float32,
            )
            for mtx in matrices
        ],
        axis=0,
    )
    rng = np.random.default_rng(seed=seed)
    sino += rng.normal(0, noise, sino.shape).astype(np.float32)
    out = da.stack(
        [
            da.from_delayed(
                iradon(sino[:, i].T, degrees, img.shape[:2]),
                shape=img.shape[:2],
                dtype=np.float32,
            )
            for i in range(img.shape[2])
        ],
        axis=1,
    )
    return out


# Radon transform


def normalize_radon_input(
    shape: tuple[int, int, int],
    central_axis: NDArray[np.float32],
    degrees: NDArray[np.float32],
):
    radians = np.deg2rad(list(degrees))

    # normalize central axis to a 3D vector
    central_axis = np.asarray(central_axis)
    central_axis /= np.sqrt(np.sum(central_axis**2))  # normalize
    if central_axis.shape != (3,):
        raise ValueError("Central axis must be a 3D vector")

    # construct Affine transform matrices
    height = int(np.ceil(np.linalg.norm(shape)))
    output_shape = (height, shape[1], shape[2])
    params = _get_rotation_matrices_for_radon_3d(
        radians, central_axis, shape, output_shape
    )

    return params, output_shape


@delayed
def radon_single(img: np.ndarray, mtx: np.ndarray, order: int = 3, output_shape=None):
    """Radon transform of 2D image."""
    img_rot = ndi.affine_transform(
        img, mtx, order=order, output_shape=output_shape, prefilter=False
    )
    return np.sum(img_rot, axis=0)


def _get_rotation_matrices_for_radon_3d(
    radians: NDArray[np.float32],
    central_axis: np.ndarray,
    in_shape: tuple[int, int, int],
    out_shape: tuple[int, int, int],
) -> Iterable[np.ndarray]:
    vec = np.stack([central_axis * rad for rad in radians], axis=0)
    in_center = (np.array(in_shape) - 1.0) / 2.0
    out_center = (np.array(out_shape) - 1.0) / 2.0
    return _utils.compose_matrices(in_center, Rotation.from_rotvec(vec), out_center)


# This function is mostly ported from `skimage.transform`.
# The most important difference is that this implementation support arbitrary
# output shape.
@delayed
def iradon(
    img: np.ndarray,
    degrees: np.ndarray,
    output_shape: tuple[int, int],
    interpolation: str = "cubic",
):
    from scipy.interpolate import interp1d

    angles_count = len(degrees)
    dtype = img.dtype
    img_shape = img.shape[0]

    # Apply filter in Fourier domain
    projection = fftn(img, axes=(0,)) * get_hamming_filter(img.shape[0])
    radon_filtered = np.real(ifftn(projection, axes=(0,))[:img_shape, :])

    # Reconstruct image by interpolation
    reconstructed = np.zeros(output_shape, dtype=dtype)
    npr, ypr = np.indices(output_shape)
    npr -= output_shape[0] // 2
    ypr -= output_shape[1] // 2

    x = np.arange(img_shape) - img_shape // 2
    for col, angle in zip(radon_filtered.T, np.deg2rad(degrees)):
        t = ypr * np.cos(angle) - npr * np.sin(angle)
        interpolant = interp1d(
            x, col, kind=interpolation, bounds_error=False, fill_value=0
        )
        reconstructed += interpolant(t)

    return reconstructed * np.pi / (2 * angles_count)


# This function is almost ported from `skimage.transform`.
def get_hamming_filter(size: int):
    n = np.concatenate(
        [
            np.arange(1, size / 2 + 1, 2, dtype=int),
            np.arange(size / 2 - 1, 0, -2, dtype=int),
        ]
    )
    f = np.zeros(size)
    f[0] = 0.25
    f[1::2] = -1 / (np.pi * n[: len(f[1::2])]) ** 2
    fourier_filter = 2 * np.real(np.fft.fft(f))  # ramp filter
    fourier_filter *= np.fft.fftshift(np.hamming(size))
    return fourier_filter[:, np.newaxis]



from __future__ import annotations

from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    Iterator,
    Literal,
    Generic,
    Sequence,
    TypeVar,
    overload,
)
import numpy as np
from numpy.typing import NDArray
from . import _bandpass, _missing_wedge

from acryo._types import degree
from scipy.spatial.transform import Rotation

_T = TypeVar("_T", bound=np.generic)
_T1 = TypeVar("_T1", bound=np.generic)

# fmt: off
class AnyArray(Generic[_T]):
    """
    Type representing a ndarray of numpy or cupy (or any other array
    that has similar API).
    """
    def __pos__(self) -> AnyArray[_T]: ...
    def __neg__(self) -> AnyArray[_T]: ...
    def __invert__(self) -> AnyArray[_T]: ...
    def __add__(self, other: Any) -> AnyArray[_T]: ...  # type: ignore
    def __sub__(self, other: Any) -> AnyArray[_T]: ...  # type: ignore
    def __mul__(self, other: Any) -> AnyArray[_T]: ...  # type: ignore
    def __truediv__(self, other: Any) -> AnyArray[np.float_]: ...  # type: ignore
    def __radd__(self, other: Any) -> AnyArray[_T]: ...  # type: ignore
    def __rsub__(self, other: Any) -> AnyArray[_T]: ...  # type: ignore
    def __rmul__(self, other: Any) -> AnyArray[_T]: ...  # type: ignore
    def __rtruediv__(self, other: Any) -> AnyArray[np.float_]: ...  # type: ignore
    def __floordiv__(self, other: Any) -> AnyArray[np.intp]: ...  # type: ignore
    def __gt__(self, other: AnyArray[_T] | float) -> AnyArray[_T]: ...  # type: ignore
    def __lt__(self, other: AnyArray[_T] | float) -> AnyArray[_T]: ...  # type: ignore
    def __ge__(self, other: AnyArray[_T] | float) -> AnyArray[_T]: ...  # type: ignore
    def __le__(self, other: AnyArray[_T] | float) -> AnyArray[_T]: ...  # type: ignore
    def __eq__(self, other: AnyArray[_T] | float) -> AnyArray[_T]: ...  # type: ignore
    def __pow__(self, other: AnyArray[_T] | float) -> AnyArray[_T]: ...  # type: ignore
    def __getitem__(self, key) -> AnyArray[_T]: ...  # type: ignore
    def __setitem__(self, key, value) -> None: ...  # type: ignore
    def __iter__(self) -> Iterator[AnyArray[_T]]: ...  # type: ignore
    @property
    def real(self) -> AnyArray[np.float32]: ...  # type: ignore
    @property
    def imag(self) -> AnyArray[np.float32]: ...  # type: ignore
    def conj(self) -> AnyArray[_T]: ...  # type: ignore
    @property
    def shape(self) -> tuple[int, ...]: ...  # type: ignore
    @property
    def ndim(self) -> int: ...
    @property
    def dtype(self) -> np.dtype[_T]: ...  # type: ignore
    def dot(self, other: AnyArray[_T]) -> AnyArray[_T]: ...  # type: ignore
    def astype(self, dtype: type[_T1]) -> AnyArray[_T1]: ...  # type: ignore
    @overload
    def mean(self, axis: None = None) -> _T: ...  # type: ignore
    @overload
    def mean(self, axis: int | tuple[int, ...]) -> AnyArray[_T]: ...  # type: ignore
    @overload
    def max(self, axis: None = None) -> _T: ...  # type: ignore
    @overload
    def max(self, axis: int | tuple[int, ...]) -> AnyArray[_T]: ...  # type: ignore
# fmt: on


class Backend:
    _default = "numpy"

    def __init__(self, name: str | None = None) -> None:
        if name is None:
            name = self._default
        if name == "numpy":
            from scipy import ndimage, fft  # type: ignore

            self._xp_ = np
            self._ndi_ = ndimage
            self._fft_ = fft
        elif name == "cupy":
            import cupy
            from cupyx.scipy import ndimage, fft

            self._xp_ = cupy
            self._ndi_ = ndimage
            self._fft_ = fft
        else:
            raise ValueError(f"Unknown backend {name}")

    @property
    def name(self) -> str:
        return self._xp_.__name__

    def __hash__(self) -> int:
        """Hash using the backend module."""
        return hash(self._xp_)

    def __repr__(self) -> str:
        return f"Backend<{self.name}>"

    def asnumpy(self, x: AnyArray[_T] | NDArray[_T]) -> NDArray[_T]:
        """Convert to numpy array."""
        if self._xp_ is np:
            return x  # type: ignore
        return x.get()  # type: ignore

    def maycopy(self, x: AnyArray[_T]) -> AnyArray[_T]:
        if self._xp_ is np:
            return x
        return x.copy()  # type: ignore

    @overload
    def array(self, x, dtype: type[_T] | np.dtype[_T]) -> AnyArray[_T]:
        ...

    @overload
    def array(self, x: AnyArray[_T] | NDArray[_T], dtype: None = None) -> AnyArray[_T]:
        ...

    def array(self, x, dtype=None):  # type: ignore
        """Convert to numpy array."""
        return self._xp_.array(x, dtype)  # type: ignore

    @overload
    def asarray(self, x, dtype: type[_T] | np.dtype[_T]) -> AnyArray[_T]:
        ...

    @overload
    def asarray(
        self, x: AnyArray[_T] | NDArray[_T], dtype: None = None
    ) -> AnyArray[_T]:
        ...

    def asarray(self, x, dtype=None):  # type: ignore
        """Convert to numpy array."""
        return self._xp_.asarray(x, dtype)  # type: ignore

    @overload
    def arange(self, *args, dtype: type[_T], **kwargs) -> AnyArray[_T]:
        ...

    @overload
    def arange(self, *args, dtype: None = None, **kwargs) -> AnyArray:
        ...

    def arange(self, *args, dtype=None, **kwargs):
        """Return evenly spaced values within a given interval."""
        return self._xp_.arange(*args, dtype=dtype, **kwargs)  # type: ignore

    @overload
    def linspace(
        self, start, stop, num: int, endpoint: bool, dtype: None = None
    ) -> AnyArray:
        ...

    @overload
    def linspace(
        self, start, stop, num: int, endpoint: bool, dtype: type[_T]
    ) -> AnyArray[_T]:
        ...

    def linspace(self, start, stop, num=50, endpoint=True, dtype=None):
        """Return evenly spaced numbers over a specified interval."""
        return self._xp_.linspace(start, stop, num, endpoint=endpoint, dtype=dtype)

    def zeros(
        self, shape: int | tuple[int, ...], dtype: type[_T] | np.dtype[_T] | None = None
    ) -> AnyArray[_T]:
        """Return a new array of given shape and type, filled with zeros."""
        return self._xp_.zeros(shape, dtype)  # type: ignore

    def full(
        self,
        shape: int | tuple[int, ...],
        fill_value: Any,
        dtype: type[_T] | np.dtype[_T] | None = None,
    ) -> AnyArray[_T]:
        """Return a new array of given shape and type, filled with fill_value."""
        return self._xp_.full(shape, fill_value, dtype=dtype)  # type: ignore

    @overload
    def sum(self, x: AnyArray[_T], axis: None = None) -> _T:
        ...

    @overload
    def sum(self, x: AnyArray[_T], axis: int | tuple[int, ...]) -> AnyArray[_T]:
        ...

    def sum(self, x, axis=None):
        """Return the sum of array elements over a given axis."""
        return self._xp_.sum(x, axis=axis)

    @overload
    def mean(self, x: AnyArray[_T], axis: None = None) -> _T:
        ...

    @overload
    def mean(self, x: AnyArray[_T], axis: int | tuple[int, ...]) -> AnyArray[_T]:
        ...

    def mean(self, x, axis=None):
        """Return the mean of array elements over a given axis."""
        return self._xp_.mean(x, axis=axis)

    def cumsum(self, x: AnyArray[_T], axis: int | None = None) -> AnyArray[_T]:
        return self._xp_.cumsum(x, axis=axis)  # type: ignore

    def sqrt(self, x: AnyArray[_T]) -> AnyArray[_T]:
        """Return the non-negative square-root of an array."""
        return self._xp_.sqrt(x)  # type: ignore

    def exp(self, x: AnyArray[_T]) -> AnyArray[_T]:
        """Return the exponential of an array."""
        return self._xp_.exp(x)  # type: ignore

    def pad(
        self,
        x: AnyArray[_T],
        pad_width: int | Sequence[int] | Sequence[tuple[int, int]],
        mode: str = "constant",
        **kwargs,
    ) -> AnyArray[_T]:
        """Pad an array."""
        return self._xp_.pad(x, pad_width, mode=mode, **kwargs)  # type: ignore

    def tensordot(
        self, a: AnyArray[_T], b: AnyArray[_T], axes: int | tuple[int, ...] = 2
    ) -> AnyArray[_T]:
        """Return tensor dot product of two arrays."""
        return self._xp_.tensordot(a, b, axes)  # type: ignore

    @overload
    def max(self, x: AnyArray[_T], axis: None = None) -> _T:
        ...

    @overload
    def max(self, x: AnyArray[_T], axis: int | tuple[int, ...]) -> AnyArray[_T]:
        ...

    def max(self, x, axis=None):
        """Return the maximum of an array or maximum along an axis."""
        return self._xp_.max(x, axis=axis)

    @overload
    def min(self, x: AnyArray[_T], axis: None = None) -> _T:
        ...

    @overload
    def min(self, x: AnyArray[_T], axis: int | tuple[int, ...]) -> AnyArray[_T]:
        ...

    def min(self, x, axis=None):
        """Return the minimum of an array or minimum along an axis."""
        return self._xp_.min(x, axis=axis)

    @overload
    def percentile(self, x: AnyArray[_T], q: float, axis: None = None) -> _T:
        ...

    @overload
    def percentile(
        self, x: AnyArray[_T], q: float, axis: int | tuple[int, ...]
    ) -> AnyArray[_T]:
        ...

    def percentile(self, x, q, axis=None):
        """Compute the q-th percentile of the data along the specified axis."""
        return self._xp_.percentile(x, q, axis=axis)

    @overload
    def argmin(self, x: AnyArray[_T], axis: None = None) -> np.intp:
        ...

    @overload
    def argmin(self, x: AnyArray[_T], axis: int | tuple[int, ...]) -> AnyArray[np.intp]:
        ...

    def argmin(self, x, axis=None):  # type: ignore
        return self._xp_.argmin(x, axis=axis)  # type: ignore

    @overload
    def argmax(self, x: AnyArray[_T], axis: None = None) -> np.intp:
        ...

    @overload
    def argmax(self, x: AnyArray[_T], axis: int | tuple[int, ...]) -> AnyArray[np.intp]:
        ...

    def argmax(self, x, axis=None):  # type: ignore
        return self._xp_.argmax(x, axis=axis)  # type: ignore

    def fix(self, x: AnyArray[_T]) -> AnyArray[_T]:
        """Round to nearest integer towards zero."""
        return self._xp_.fix(x)  # type: ignore

    def fftn(
        self,
        x: AnyArray[np.float32] | AnyArray[np.complex64],
        s: tuple[int, int, int] | None = None,
        axes: int | tuple[int, ...] | None = None,
    ) -> AnyArray[np.complex64]:
        """N-dimensional FFT."""
        return self._fft_.fftn(x, s, axes)  # type: ignore

    def ifftn(
        self,
        x: AnyArray[np.float32] | AnyArray[np.complex64],
        s: tuple[int, int, int] | None = None,
        axes: int | tuple[int, ...] | None = None,
    ) -> AnyArray[np.complex64]:
        """N-dimensional inverse FFT."""
        return self._fft_.ifftn(x, s, axes)  # type: ignore

    def rfftn(
        self,
        x: AnyArray[np.float32],
        s: tuple[int, int, int] | None = None,
        axes: int | tuple[int, ...] | None = None,
    ) -> AnyArray[np.complex64]:
        """N-dimensional FFT of real part."""
        return self._fft_.rfftn(x, s, axes)  # type: ignore

    def irfftn(
        self,
        x: AnyArray[np.complex64],
        s: tuple[int, int, int] | None = None,
        axes: int | tuple[int, ...] | None = None,
    ) -> AnyArray[np.float32]:
        """N-dimensional inverse FFT of real part."""
        return self._fft_.irfftn(x, s, axes)  # type: ignore

    def fftshift(self, x: AnyArray[_T], axes=None) -> AnyArray[_T]:
        """Shift zero-frequency component to center."""
        return self._xp_.fft.fftshift(x, axes=axes)  # type: ignore

    def ifftshift(self, x: AnyArray[_T], axes=None) -> AnyArray[_T]:
        """Inverse shift zero-frequency component to center."""
        return self._xp_.fft.ifftshift(x, axes=axes)  # type: ignore

    def fftfreq(self, n: int, d: float = 1.0) -> AnyArray[np.float_]:
        """Return the Discrete Fourier Transform sample frequencies."""
        return self._xp_.fft.fftfreq(n, d)  # type: ignore

    def meshgrid(
        self,
        *xi: AnyArray[_T],
        copy: bool = True,
        sparse: bool = False,
        indexing: Literal["xy", "ij"] = "xy",
    ) -> tuple[AnyArray[_T], ...]:
        """Return coordinate matrices from coordinate vectors."""
        return self._xp_.meshgrid(*xi, copy=copy, sparse=sparse, indexing=indexing)  # type: ignore

    @overload
    def indices(
        self, shape: tuple[int], dtype: type[_T] = np.int32
    ) -> tuple[AnyArray[_T]]:
        ...

    @overload
    def indices(
        self, shape: tuple[int, int], dtype: type[_T] = np.int32
    ) -> tuple[AnyArray[_T], AnyArray[_T]]:
        ...

    @overload
    def indices(
        self, shape: tuple[int, int, int], dtype: type[_T] = np.int32
    ) -> tuple[AnyArray[_T], AnyArray[_T], AnyArray[_T]]:
        ...

    @overload
    def indices(
        self, shape: tuple[int, ...], dtype: type[_T] = np.int32
    ) -> tuple[AnyArray[_T], ...]:
        ...

    def indices(self, shape, dtype=np.int32):  # type: ignore
        """Return an array representing the indices of a grid."""
        return self._xp_.indices(shape, dtype=dtype)  # type: ignore

    def unravel_index(self, indices, shape: tuple[int, ...]) -> AnyArray[np.intp]:
        """Converts a flat index into a tuple of coordinate arrays."""
        return self._xp_.asarray(self._xp_.unravel_index(indices, shape))  # type: ignore

    def stack(self, arrays: Sequence[AnyArray[_T]], axis: int = 0) -> AnyArray[_T]:
        """Stack arrays in sequence along a new axis."""
        return self._xp_.stack(arrays, axis=axis)  # type: ignore

    def affine_transform(
        self,
        img,
        matrix,
        output_shape: tuple[int, ...] | None = None,
        output=None,
        order: int = 3,
        mode: str = "constant",
        cval: float = 0.0,
        prefilter: bool = True,
    ) -> AnyArray[np.float32]:
        """Affine transform."""
        return self._ndi_.affine_transform(
            self.asarray(img),
            self.asarray(matrix),
            output_shape=output_shape,
            output=output,
            order=order,
            mode=mode,
            cval=float(cval),
            prefilter=prefilter,
        )  # type: ignore

    def spline_filter(
        self,
        input,
        order: int = 3,
        output: type[_T] = np.float64,
        mode: str = "mirror",
    ) -> AnyArray[_T]:
        return self._ndi_.spline_filter(
            self.asarray(input), order=order, output=output, mode=mode  # type: ignore
        )

    def map_coordinates(
        self,
        x: AnyArray[_T],
        coords: AnyArray[_T],
        order: int = 3,
        mode: str = "constant",
        cval: float = -1.0,
        prefilter: bool = True,
    ) -> AnyArray[_T]:
        return self._ndi_.map_coordinates(
            x, coords, order=order, mode=mode, cval=cval, prefilter=prefilter
        )  # type: ignore

    def sum_labels(
        self,
        arr: AnyArray[_T],
        labels: AnyArray[np.uint16],
        index: AnyArray[np.uint16],
    ) -> AnyArray[np.uint16]:
        return self._ndi_.sum_labels(arr, labels=labels, index=index)  # type: ignore

    def rotated_crop(
        self,
        subimg,
        mtx: NDArray[np.float32],
        shape: tuple[int, int, int],
        order: int,
        cval: float | Callable[[AnyArray[np.float32]], Any],
    ) -> AnyArray[np.float32]:
        if callable(cval):
            _cval = cval(subimg)
        else:
            _cval = cval

        out = self.affine_transform(
            subimg,
            matrix=self.asarray(mtx),
            output_shape=shape,
            order=order,
            prefilter=order > 1,
            mode="constant",
            cval=float(_cval),
        )
        return out

    def lowpass_filter_ft(self, img, cutoff, order: int = 2) -> AnyArray[np.complex64]:
        """Lowpass filter in Fourier space."""
        return _bandpass.lowpass_filter_ft(self, self.asarray(img), cutoff, order)

    def lowpass_filter(self, img, cutoff, order: int = 2) -> AnyArray[np.float32]:
        """Lowpass filter in real space."""
        return _bandpass.lowpass_filter(self, self.asarray(img), cutoff, order)

    def missing_wedge_mask(
        self,
        rotator: Rotation,
        tilt_range: tuple[degree, degree],
        shape: tuple[int, int, int],
    ):
        return _missing_wedge.missing_wedge_mask(self, rotator, tilt_range, shape)


NUMPY_BACKEND = Backend("numpy")


@contextmanager
def using_backend(name: str):
    """Context manager to temporarily change the default backend."""
    old_backend = Backend._default
    Backend._default = name
    try:
        yield
    finally:
        Backend._default = old_backend


def set_backend(name: str):
    """Set the default backend."""
    Backend._default = name



# pyright: reportPrivateImportUsage=false
from __future__ import annotations
from typing import Any

import numpy as np
from numpy.typing import NDArray
from dask import array as da
from scipy import ndimage as ndi

from acryo.pick._base import BasePickerModel, BaseTemplateMatcher
from acryo.backend import NUMPY_BACKEND
from acryo.backend._zncc import ncc_landscape_no_pad
from acryo.molecules import Molecules
from acryo._types import nm


class ZNCCTemplateMatcher(BaseTemplateMatcher):
    """
    Particle picking based on ZNCC template matching.

    Parameters
    ----------
    image : da.Array
        The input image.
    scale : float
        The scale of the image.
    min_distance : float
        The minimum distance between the picked particles.
    min_score : float
        The minimum score of the picked particles.
    boundary : str
        The boundary condition for the template matching.
    """

    def pick_molecules(
        self,
        image: da.Array,
        scale: nm = 1.0,
        *,
        min_distance: nm = 1.0,
        min_score: float = 0.02,
        boundary="nearest",
    ) -> Molecules:
        return super().pick_molecules(
            image,
            scale,
            boundary=boundary,
            min_distance=min_distance / scale,
            min_score=min_score,
        )

    def pick_in_chunk(
        self,
        image: NDArray[np.float32],
        templates: list[NDArray[np.float32]],
        min_distance: float,  # pixel
        min_score: float,
    ):
        all_landscapes = np.stack(
            [
                ncc_landscape_no_pad(
                    image - np.mean(image),
                    template - np.mean(template),
                    NUMPY_BACKEND,
                )
                for template in templates
            ],
            axis=0,
        )

        img_argmax = np.argmax(all_landscapes, axis=0)
        landscale_max = np.max(all_landscapes, axis=0)

        pos = find_maxima(landscale_max, min_distance, min_score)
        argmax_indices = np.array(
            [img_argmax[tuple(np.round(p).astype(np.int32))] for p in pos]
        )
        score = _sample_score(landscale_max, pos)
        quats = self._index_to_quaternions(argmax_indices)
        offset = (np.array(templates[0].shape) + 1) / 2
        return pos + offset, quats, {"score": score}


class LoGPicker(BasePickerModel):
    """Particle picking based on Laplacian of Gaussian."""

    def __init__(self, sigma: nm = 3.5) -> None:
        self._sigma = sigma

    def pick_in_chunk(
        self,
        image: NDArray[np.float32],
        sigma: float,
    ) -> tuple[NDArray[np.float32], NDArray[np.uint16], Any]:
        img_filt = -ndi.gaussian_laplace(image, sigma)
        pos = find_maxima(img_filt, sigma, 0.0)
        return simple_pick(img_filt, pos)

    def get_params_and_depth(self, scale: nm):
        sigma_px = self._sigma / scale
        depth = int(np.ceil(sigma_px * 2))
        return {"sigma": sigma_px}, depth


class DoGPicker(BasePickerModel):
    """Particle picking based on Difference of Gaussian."""

    def __init__(self, sigma_low: nm = 3.5, sigma_high: nm = 5.0) -> None:
        if sigma_low >= sigma_high:
            raise ValueError("sigma_low must be smaller than sigma_high")
        self._sigma_low = sigma_low
        self._sigma_high = sigma_high

    def pick_in_chunk(
        self,
        image: NDArray[np.float32],
        sigma_low: float,
        sigma_high: float,
    ) -> tuple[NDArray[np.float32], NDArray[np.uint16], Any]:
        img_filt = _differece_of_gaussian(image, sigma_low, sigma_high)
        pos = find_maxima(img_filt, sigma_low, 0.0)
        return simple_pick(img_filt, pos)

    def get_params_and_depth(self, scale: nm):
        sigma1_px = self._sigma_low / scale
        sigma2_px = self._sigma_high / scale
        depth = int(np.ceil(sigma1_px * 2))
        return {"sigma_low": sigma1_px, "sigma_high": sigma2_px}, depth


def maximum_filter(image: da.Array, radius: float) -> NDArray[np.float32]:
    if radius < 1:
        return image
    r_int = int(np.ceil(radius))
    size = 2 * r_int + 1
    zz, yy, xx = np.indices((size,) * 3)
    foot = (zz - r_int) ** 2 + (yy - r_int) ** 2 + (xx - r_int) ** 2 <= radius**2
    return ndi.maximum_filter(image, footprint=foot, mode="nearest")  # type: ignore


def find_maxima(img, min_distance: float, min_intensity: float):
    img_max_maxfilt = maximum_filter(img, min_distance)
    is_maxima = (img_max_maxfilt == img) & (img > min_intensity)
    s0 = [[0, 0, 0], [0, 1, 0], [0, 0, 0]]
    s1 = [[0, 1, 0], [1, 1, 1], [0, 1, 0]]
    structure = np.stack([s0, s1, s0])
    label_img, nfeat = ndi.label(is_maxima, structure=structure)
    centers = ndi.center_of_mass(img, label_img, range(1, nfeat + 1))
    return np.array(centers, dtype=np.float32)


def simple_pick(img: NDArray[np.float32], pos: NDArray[np.float32]):
    score = _sample_score(img, pos)
    quats = np.zeros((pos.shape[0], 4), dtype=np.float32)
    quats[:, 3] = 1.0
    return pos, quats, {"score": score}


def _sample_score(img, pos: NDArray[np.float32]) -> NDArray[np.float32]:
    return ndi.map_coordinates(img, pos.T, order=3, mode="reflect")


def _differece_of_gaussian(
    image: NDArray[np.float32],
    sigma_low: float,
    sigma_high: float,
) -> NDArray[np.float32]:
    img_l = ndi.gaussian_filter(image, sigma_low)
    img_h = ndi.gaussian_filter(image, sigma_high)
    return img_l - img_h



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from ._base import TomographyInput
from acryo._types import pixel
from acryo.backend import Backend, AnyArray
from acryo.backend._pcc import subpixel_pcc, pcc_landscape
from acryo.backend._zncc import subpixel_zncc, zncc_landscape_with_crop
from acryo.backend._fsc import subpixel_fsc, fsc_landscape


class PCCAlignment(TomographyInput):
    """Alignment model using phase cross correlation."""

    def _optimize(
        self,
        subvolume: AnyArray[np.complex64],
        template: AnyArray[np.complex64],
        max_shifts: tuple[pixel, pixel, pixel],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> tuple[NDArray[np.float32], NDArray[np.float32], float]:
        """Optimize."""
        mw = self._get_missing_wedge_mask(quaternion, backend)
        shift, pcc = subpixel_pcc(
            subvolume * mw,
            template * mw,
            upsample_factor=20,
            max_shifts=max_shifts,
            backend=backend,
        )
        return shift, self._DUMMY_QUAT, pcc

    def _landscape(
        self,
        subvolume: AnyArray[np.complex64],
        template: AnyArray[np.complex64],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> AnyArray[np.float32]:
        """Compute landscape."""
        mw = self._get_missing_wedge_mask(quaternion, backend)
        return pcc_landscape(
            subvolume * mw,
            template * mw,
            max_shifts=max_shifts,
            backend=backend,
        )


class ZNCCAlignment(TomographyInput):
    """Alignment model using zero-mean normalized cross correlation."""

    def _optimize(
        self,
        subvolume: AnyArray[np.complex64],
        template: AnyArray[np.complex64],
        max_shifts: tuple[pixel, pixel, pixel],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> tuple[NDArray[np.float32], NDArray[np.float32], float]:
        """Optimize."""
        mw = self._get_missing_wedge_mask(quaternion, backend)
        shift, zncc = subpixel_zncc(
            backend.ifftn(subvolume * mw).real,
            backend.ifftn(template * mw).real,
            max_shifts=max_shifts,
            backend=backend,
        )
        return shift, self._DUMMY_QUAT, zncc

    def _landscape(
        self,
        subvolume: AnyArray[np.complex64],
        template: AnyArray[np.complex64],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> AnyArray[np.float32]:
        """Compute landscape."""
        mw = self._get_missing_wedge_mask(quaternion, backend)
        return zncc_landscape_with_crop(
            backend.ifftn(subvolume * mw).real,
            backend.ifftn(template * mw).real,
            max_shifts=max_shifts,
            backend=backend,
        )


class FSCAlignment(TomographyInput):
    """Alignment model using Fourier shell correlation."""

    def _optimize(
        self,
        subvolume: AnyArray[np.complex64],
        template: AnyArray[np.complex64],
        max_shifts: tuple[pixel, pixel, pixel],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> tuple[NDArray[np.float32], NDArray[np.float32], float]:
        """Optimize."""
        mw = self._get_missing_wedge_mask(quaternion, backend)
        shift, fsc = subpixel_fsc(
            subvolume * mw,
            template * mw,
            max_shifts=max_shifts,
            backend=backend,
        )
        return shift, self._DUMMY_QUAT, fsc

    def _landscape(
        self,
        subvolume: AnyArray[np.complex64],
        template: AnyArray[np.complex64],
        max_shifts: tuple[float, float, float],
        quaternion: NDArray[np.float32],
        pos: NDArray[np.float32],
        backend: Backend,
    ) -> AnyArray[np.float32]:
        """Compute landscape."""
        mw = self._get_missing_wedge_mask(quaternion, backend)
        return fsc_landscape(
            subvolume * mw,
            template * mw,
            max_shifts=max_shifts,
            backend=backend,
        )



__version__ = "0.4.11"

from acryo.loader import SubtomogramLoader, BatchLoader, MockLoader
from acryo.molecules import Molecules
from acryo.simulator import TomogramSimulator

imread = SubtomogramLoader.imread

__all__ = [
    "Molecules",
    "SubtomogramLoader",
    "BatchLoader",
    "MockLoader",
    "TomogramSimulator",
    "imread",
]



from __future__ import annotations

from functools import lru_cache
import math
import numpy as np
from numpy.typing import NDArray


@lru_cache(maxsize=8)
def get_norms_y(
    tilt_range: tuple[float, float]
) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
    degmin, degmax = tilt_range
    ang0 = math.pi - math.radians(degmin)
    ang1 = math.pi - math.radians(degmax)
    return (
        np.array([math.cos(ang0), 0.0, math.sin(ang0)], dtype=np.float32),
        np.array([math.cos(ang1), 0.0, math.sin(ang1)], dtype=np.float32),
    )


@lru_cache(maxsize=8)
def get_norms_x(
    tilt_range: tuple[float, float]
) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
    degmin, degmax = tilt_range
    ang0 = math.pi - math.radians(degmin)
    ang1 = math.pi - math.radians(degmax)
    return (
        np.array([math.cos(ang0), math.sin(ang0), 0.0], dtype=np.float32),
        np.array([math.cos(ang1), math.sin(ang1), 0.0], dtype=np.float32),
    )


@lru_cache(maxsize=32)
def get_indices(shape: tuple[int, int, int]) -> NDArray[np.float32]:
    inds = np.indices(shape, dtype=np.float32)
    for ind, s in zip(inds, shape):
        # Note that the shifts in indices must resemble the shifts in fftshift.
        ind -= math.ceil(s / 2)
    return np.fft.fftshift(np.stack(list(inds), axis=-1), axes=(0, 1, 2))



# pyright: reportPrivateImportUsage=false

from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    NamedTuple,
    Any,
)
import numpy as np
from dask import array as da

from acryo._types import nm, pixel
from acryo._reader import imread
from acryo.molecules import Molecules
from acryo.backend import Backend
from acryo import _utils
from acryo.loader import _misc
from acryo.loader._base import LoaderBase, Unset, _ShapeType
from acryo._dask import DaskTaskPool, DaskArrayList

if TYPE_CHECKING:
    from typing_extensions import Self
    from numpy.typing import NDArray
    from acryo.classification import PcaClassifier


class SubtomogramLoader(LoaderBase):
    """
    A class for efficient loading of subtomograms.

    A ``SubtomogramLoader`` instance is basically composed of two elements,
    an image and a Molecules object. A subtomogram is loaded by creating a
    local rotated Cartesian coordinate at a molecule and calculating mapping
    from the image to the subtomogram.

    Parameters
    ----------
    image : np.ndarray or da.Array
        Tomogram image. Must be 3-D.
    molecules : Molecules
        Molecules object that represents positions and orientations of
        subtomograms.
    order : int, default is 3
        Interpolation order of subtomogram sampling.
        - 0 = Nearest neighbor
        - 1 = Linear interpolation
        - 3 = Cubic interpolation
    scale : float, default is 1.0
        Physical scale of pixel, such as nm. This value does not affect
        averaging/alignment results but molecule coordinates are multiplied
        by this value. This parameter is useful when another loader with
        binned image is created.
    output_shape : int or tuple of int, optional
        Shape of output subtomogram in pixel. This parameter is not required
        if template (or mask) image is available immediately.
    corner_safe : bool, default is False
        If true, regions around molecules will be cropped at a volume larger
        than ``output_shape`` so that densities at the corners will not be
        lost due to rotation. If target density is globular, this parameter
        should be set false to save computation time.
    """

    def __init__(
        self,
        image: np.ndarray | da.Array,
        molecules: Molecules,
        order: int = 3,
        scale: nm = 1.0,
        output_shape: pixel | tuple[pixel, pixel, pixel] | Unset = Unset(),
        corner_safe: bool = False,
    ) -> None:
        # check type of input image
        if not isinstance(image, (np.ndarray, da.Array)):
            raise TypeError(
                "Input image of a SubtomogramLoader instance must be np.ndarray "
                f"or dask.Array, got {type(image)}."
            )

        self._image = image

        # check type of molecules
        if not isinstance(molecules, Molecules):
            raise TypeError(
                "The second argument 'molecules' must be a Molecules object, got"
                f"{type(molecules)}."
            )
        self._molecules = molecules

        super().__init__(
            order=order, scale=scale, output_shape=output_shape, corner_safe=corner_safe
        )

    def __repr__(self) -> str:
        shape = self.image.shape
        mole_repr = repr(self.molecules)
        return (
            f"{self.__class__.__name__}(tomogram={shape}, molecules={mole_repr}, "
            f"output_shape={self.output_shape}, order={self.order}, "
            f"scale={self.scale:.4f})"
        )

    @classmethod
    def imread(
        cls,
        path: str,
        molecules: Molecules,
        order: int = 3,
        scale: nm | None = None,
        output_shape: pixel | tuple[pixel, pixel, pixel] | Unset = Unset(),
        corner_safe: bool = False,
        chunks: Any = "auto",
    ):
        dask_array, _scale = imread(str(path), chunks)
        if scale is None:
            scale = _scale
        return cls(
            dask_array,
            molecules=molecules,
            order=order,
            scale=scale,
            output_shape=output_shape,
            corner_safe=corner_safe,
        )

    @property
    def image(self) -> NDArray[np.float32] | da.Array:
        """Return tomogram image."""
        return self._image

    @property
    def molecules(self) -> Molecules:
        """Return the molecules of the subtomogram loader."""
        return self._molecules

    def __len__(self) -> int:
        """Return the number of subtomograms."""
        return self.molecules.pos.shape[0]

    def replace(
        self,
        molecules: Molecules | None = None,
        output_shape: pixel | tuple[pixel, pixel, pixel] | Unset | None = None,
        order: int | None = None,
        scale: float | None = None,
        corner_safe: bool | None = None,
    ) -> Self:
        """Return a new instance with different parameter(s)."""
        if molecules is None:
            molecules = self.molecules
        if output_shape is None:
            output_shape = self.output_shape
        if order is None:
            order = self.order
        if scale is None:
            scale = self.scale
        if corner_safe is None:
            corner_safe = self.corner_safe
        return self.__class__(
            self.image,
            molecules=molecules,
            output_shape=output_shape,
            order=order,
            scale=scale,
            corner_safe=corner_safe,
        )

    def binning(self, binsize: pixel = 2, *, compute: bool = True) -> Self:
        """
        Return a new instance with binned image.

        This method also properly translates the molecule coordinates.

        Parameters
        ----------
        binsize : int, default is 2
            Bin size.
        compute : bool, default is True
            If true, the image is computed immediately to a numpy array.

        Returns
        -------
        SubtomogramLoader
            A new instance with binned image.
        """
        if binsize == 1:
            return self.copy()
        tr = -(binsize - 1) / 2 * self.scale
        molecules = self.molecules.translate([tr, tr, tr])
        binned_image = _utils.bin_image(self.image, binsize=binsize)
        if isinstance(binned_image, da.Array) and compute:
            binned_image = binned_image.compute()
        out = self.replace(
            molecules=molecules,
            scale=self.scale * binsize,
        )

        out._image = binned_image
        return out

    def construct_loading_tasks(
        self,
        output_shape: _ShapeType = None,
        backend: Backend | None = None,
    ) -> DaskArrayList:
        """
        Construct a list of subtomogram lazy loader.

        Returns
        -------
        list of Delayed object
            Each object returns a subtomogram on execution by ``da.compute``.
        """
        output_shape = self._get_output_shape(output_shape)
        xp = backend or Backend()

        image = self.image
        scale = self.scale
        if isinstance(image, np.ndarray):
            image = da.from_array(image, asarray=xp.asarray)

        if self.corner_safe:
            _prep = _utils.prepare_affine_cornersafe
        else:
            _prep = _utils.prepare_affine
        pool = DaskTaskPool.from_func(xp.rotated_crop)
        for i in range(self.molecules.count()):
            try:
                subvol, mtx = _prep(
                    image,
                    center=self.molecules.pos[i] / scale,
                    output_shape=output_shape,
                    rot=self.molecules.rotator[i],
                    order=self.order,
                )
            except _utils.SubvolumeOutOfBoundError as err:
                raise err.with_msg(
                    f"The {i}-th molecule at {tuple(self.molecules.pos[i])} is "
                    f"out of bound. {err.msg}"
                )
            pool.add_task(
                subvol,
                mtx,
                shape=output_shape,
                order=self.order,
                cval=xp.mean,
            )

        return pool.asarrays(shape=output_shape, dtype=np.float32)


class ClassificationResult(NamedTuple):
    """Tuple of classification results."""

    loader: SubtomogramLoader
    classifier: PcaClassifier


def check_input(
    order: int,
    output_shape: pixel | tuple[pixel, pixel, pixel] | Unset,
    scale: float,
    corner_safe: bool,
    ndim: int,
):
    # check interpolation order
    if order not in (0, 1, 3):
        raise ValueError(
            f"The third argument 'order' must be 0, 1 or 3, got {order!r}."
        )

    # check output_shape
    if isinstance(output_shape, Unset):
        _output_shape = output_shape
    else:
        _output_shape = _misc.normalize_shape(output_shape, ndim=ndim)

    # check scale
    _scale = float(scale)
    if _scale <= 0:
        raise ValueError("Negative scale is not allowed.")

    _corner_safe = bool(corner_safe)
    return order, _output_shape, _scale, _corner_safe



from __future__ import annotations

from typing import Callable, Any, TypeVar
from typing_extensions import ParamSpec, Concatenate
import inspect
import numpy as np
from acryo.pipe._classes import ImageProvider, ImageConverter
from acryo._types import nm

_P = ParamSpec("_P")
_R = TypeVar("_R", np.ndarray, "list[np.ndarray]")


def provider_function(
    fn: Callable[Concatenate[nm | float, _P], _R]
) -> Callable[_P, ImageProvider[_R]]:
    """
    Convert a function into a curried function that returns a image provider.

    Examples
    --------
    >>> @provider_function
    ... def provide_random_image(scale, shape):
    ...     return np.random.random(shape)
    >>> provider = provide_random_image((10, 20, 30))
    >>> provider(0.18)  # return a (10, 20, 30) array

    """

    def inner(*args: _P.args, **kwargs: _P.kwargs):
        _fn = _assert_1_arg(fn)
        return ImageProvider(lambda scale: _fn(scale, *args, **kwargs)).with_name(
            _format_args(fn, *args, **kwargs)
        )

    _update_wrapper(inner, fn, npop=1)
    return inner


def converter_function(
    fn: Callable[Concatenate[np.ndarray, nm | float, _P], np.ndarray]
) -> Callable[_P, ImageConverter]:
    """
    Convert a function into a curried function that returns a image converter.

    Input function must accept `fn(img, scale)`.

    Examples
    --------
    >>> from scipy import ndimage as ndi
    >>> @converter_function
    ... def gaussian_filter(img, scale, sigma):
    ...     return ndi.gaussian_filter(img, sigma / scale)
    >>> converter = gaussian_filter(1.5)
    >>> converter(arr)  # return a Gaussian filtered `arr`

    """

    def inner(*args: _P.args, **kwargs: _P.kwargs):
        _fn = _assert_2_args(fn)
        return ImageConverter(
            lambda img, scale: _fn(img, scale, *args, **kwargs)
        ).with_name(_format_args(fn, *args, **kwargs))

    _update_wrapper(inner, fn, npop=2)
    return inner


def _update_wrapper(f, wrapped, npop: int):
    _update_attr(f, wrapped)
    annot = getattr(wrapped, "__annotations__", {}).copy()
    if len(annot) == 0:
        return f
    try:
        args = inspect.getargs(wrapped.__code__)
    except Exception:
        pass
    else:
        for i in range(npop):
            name = args.args[i]
            annot.pop(name, None)
        f.__annotations__ = annot
    return f


def _assert_1_arg(func: Callable) -> Callable[[Any], Any]:
    nargs = sum(
        1
        for p in inspect.signature(func).parameters.values()
        if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD)
    )
    if nargs == 0:
        out = lambda x: func()
        _update_attr(out, func)
        return out
    else:
        return func


def _assert_2_args(func: Callable) -> Callable[[Any, Any], Any]:
    nargs = sum(
        1
        for p in inspect.signature(func).parameters.values()
        if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD)
    )
    if nargs == 0:
        out = lambda x0, x1: func()
    elif nargs == 1:
        out = lambda x0, x1: func(x0)
    else:
        return func
    _update_attr(out, func)
    return out


def _update_attr(f, wrapped):
    for name in ("__module__", "__name__", "__qualname__", "__doc__"):
        try:
            value = getattr(wrapped, name)
        except AttributeError:
            pass
        else:
            setattr(f, name, value)


# formatter
def _format(arg: Any) -> str:
    if isinstance(arg, (float, complex)):
        return format(arg, ".2f")
    elif hasattr(arg, "__array__"):
        return "..."
    else:
        return repr(arg)


def _format_args(fn: Callable, *args, **kwargs):
    _args = [_format(arg) for arg in args]
    _kwargs = [f"{k}={_format(v)}" for k, v in kwargs.items()]

    s = ", ".join(_args + _kwargs)
    return f"{fn.__name__}({s})"



from __future__ import annotations

from typing import TYPE_CHECKING
import numpy as np
from numpy.typing import NDArray
from scipy import ndimage as ndi
from acryo import _utils
from acryo._types import nm
from acryo._typed_scipy import shift as ndi_shift
from acryo.pipe._curry import converter_function

if TYPE_CHECKING:
    from acryo._typed_scipy import Mode


@converter_function
def center_by_mass(
    img: NDArray[np.float32], scale: nm, *, order: int = 3
) -> NDArray[np.float32]:
    """Centering an image by its center of mass."""
    shift = np.array(ndi.center_of_mass(img)) - np.array(img.shape) / 2
    return ndi_shift(img, -shift, order=order, prefilter=order > 1, mode="nearest")


@converter_function
def gaussian_filter(
    img: NDArray[np.float32], scale: nm, *, sigma: nm, mode="reflect", cval: float = 0.0
) -> NDArray[np.float32]:
    """Gaussian filtering an image."""
    return ndi.gaussian_filter(img, sigma / scale, mode=mode, cval=cval)  # type: ignore


@converter_function
def lowpass_filter(img: NDArray[np.float32], scale: nm, cutoff: float, order: int = 2):
    return _utils.lowpass_filter(img, cutoff=cutoff, order=order)


@converter_function
def highpass_filter(img: NDArray[np.float32], scale: nm, cutoff: float, order: int = 2):
    return _utils.highpass_filter(img, cutoff=cutoff, order=order)


@converter_function
def shift(
    img: NDArray[np.float32],
    scale: nm,
    shift: tuple[nm, nm, nm],
    *,
    order: int = 3,
    mode: Mode = "nearest",
    cval: float = 0.0,
) -> NDArray[np.float32]:
    """Shift the image by nm."""
    shift_px = np.asarray(shift) / scale
    return ndi_shift(img, shift_px, order=1, prefilter=order > 1, mode=mode, cval=cval)



from __future__ import annotations

from typing import Iterable
from functools import reduce
from abc import ABC, abstractmethod
import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation


class TiltSeriesModel(ABC):
    @abstractmethod
    def create_mask(
        self,
        rotator: Rotation,
        shape: tuple[int, int, int],
    ) -> NDArray[np.float32]:
        """Create a missing wedge mask."""

    def apply_mask(
        self,
        rotator: Rotation,
        img: NDArray[np.float32],
    ) -> NDArray[np.float32]:
        """Apply the missing wedge mask."""
        mask = self.create_mask(rotator, img.shape)
        return img * mask


class NoWedge(TiltSeriesModel):
    def create_mask(
        self,
        rotator: Rotation = Rotation.identity(),
        shape: tuple[int, int, int] = (49, 49, 49),
    ) -> NDArray[np.float32]:
        """
        Create an array filled with 1.0.

        Parameters
        ----------
        rotator : Rotation
            Does not affect the result.
        shape : tuple of int
            The shape of the mask.

        Returns
        -------
        np.ndarray
            Missing wedge mask.
        """
        return np.ones(shape, dtype=np.float32)


class UnionAxes(TiltSeriesModel):
    def __init__(self, wedges: Iterable[TiltSeriesModel]):
        self._wedges = list(wedges)

    def __repr__(self) -> str:
        reprs = ", ".join(repr(w) for w in self._wedges)
        return f"{self.__class__.__name__}({reprs})"

    def create_mask(
        self,
        rotator: Rotation = Rotation.identity(),
        shape: tuple[int, int, int] = (49, 49, 49),
    ) -> NDArray[np.float32]:
        """
        Create a binary mask that covers tomographical missing wedge.

        Note that the mask is not shifted to the center of the Fourier domain.
        ``np.fft.fftn(img) * mask`` will be the correct way to apply the mask.

        Parameters
        ----------
        rotator : Rotation
            The rotation object that describes the direction of the mask.
        shape : tuple of int
            The shape of the mask.

        Returns
        -------
        np.ndarray
            Missing wedge mask.
        """
        return reduce(np.maximum, (w.create_mask(rotator, shape) for w in self._wedges))



from ._masking import gaussian_smooth, soft_otsu, threshold_otsu, dilation, closing
from ._transform import (
    center_by_mass,
    gaussian_filter,
    lowpass_filter,
    highpass_filter,
    shift,
)
from ._classes import ImageProvider, ImageConverter
from ._curry import provider_function, converter_function
from ._imread import (
    from_file,
    from_files,
    from_gaussian,
    from_array,
    from_arrays,
    from_atoms,
    from_pdb,
)

__all__ = [
    "gaussian_smooth",
    "soft_otsu",
    "threshold_otsu",
    "dilation",
    "closing",
    "center_by_mass",
    "gaussian_filter",
    "lowpass_filter",
    "highpass_filter",
    "shift",
    "from_file",
    "from_files",
    "from_gaussian",
    "from_array",
    "from_arrays",
    "from_atoms",
    "from_pdb",
    "provider_function",
    "converter_function",
    "ImageProvider",
    "ImageConverter",
]
